from sqlalchemy import Select, Result, ScalarResult
from sqlalchemy.ext.asyncio import async_sessionmaker, AsyncSession

from app import get_logger
from app.enum import ResultStatus
from app.models.configure import Configure
from app.entity.configure import Configure as EConfig


async def get_all(db: async_sessionmaker[AsyncSession]) -> list[EConfig]:
    result: list[EConfig] = []
    res: Result = None
    sca: ScalarResult = None

    async with db() as ses:
        try:
            stmt = Select(Configure)
            res = await ses.execute(stmt)
            sca = res.scalars()
            for row in sca:
                result.append(EConfig.model_validate(row))
        except Exception:
            result = []
            get_logger(__name__).exception("設定情報リストの取得に失敗しました。")
        finally:
            if sca is not None:
                sca.close()
                sca = None
            if res is not None:
                res.close()
                res = None
            await ses.rollback()

    return result


async def save(db: async_sessionmaker[AsyncSession], ent: EConfig):
    result = None
    res: Result = None
    orm: Configure = None

    async with db() as ses:
        try:
            stmt = Select(Configure).where(Configure.ckey == ent.ckey)
            res = await ses.execute(stmt)
            sca = res.scalar_one_or_none()
            if sca is not None:
                sca.update(**ent.model_dump())
                status = ResultStatus.update
            else:
                orm = Configure(**ent.model_dump(exclude=["update_at"]))
                ses.add(orm)
                status = ResultStatus.create
            await ses.commit()
            await ses.flush()
            result = {"key": ent.ckey, "status": status}
        except Exception:
            get_logger(__name__).exception("設定の保存に失敗しました。")
            await ses.rollback()
            result = None
        finally:
            if res is not None:
                res.close()
                res = None
            orm = None

    return result


async def delete(db: async_sessionmaker[AsyncSession], ckey: str):
    result = None
    res: Result = None
    log = get_logger(__name__)
    async with db() as ses:
        try:
            stmt = Select(Configure).where(Configure.ckey == ckey)
            res = await ses.execute(stmt)
            sca = res.scalar_one_or_none()
            if sca is not None:
                await ses.delete(sca)
                await ses.commit()
                await ses.flush()

                result = {"key": ckey, "status": ResultStatus.delete}
            else:
                await ses.rollback()
                log.warning(f"キー '{ckey}' が見つかりません。")
        except Exception:
            log.exception("設定の削除に失敗しました。")
            await ses.rollback()
            result = None
        finally:
            if res is not None:
                res.close()
                res = None

    return result
