from sqlalchemy import Select, func, Result, ScalarResult
from sqlalchemy.sql.expression import func as fc
from sqlalchemy.orm import joinedload
from sqlalchemy.ext.asyncio import async_sessionmaker, AsyncSession

from app import get_logger
from app.enum import ResultStatus
from app.models.mail_filter import MailFilter
from app.entity.mail_filter import MailFilter as EMailFilter, MailFilterListItem
from app.models.users import Users


async def create(db: async_sessionmaker[AsyncSession], ent: EMailFilter) -> dict[str, any]:
    result: dict = None
    orm: MailFilter = None
    async with db() as ses:
        try:
            exc = ["filter_id", "update_at"]
            orm = MailFilter.dict_to_model(ent.model_dump(exclude=exc), exclude=exc)
            ses.add(orm)
            await ses.commit()
            await ses.flush()
            await ses.refresh(orm)
            result = {"filter_id": orm.filter_id, "status": ResultStatus.create}
        except Exception:
            get_logger(__name__).exception("メールフィルターの作成に失敗しました。")
            await ses.rollback()
            result = None
    orm = None

    return result


async def get_list(db: async_sessionmaker[AsyncSession], lmt: int = 0, off_s: int = 0) -> list[EMailFilter]:
    result: list[EMailFilter] = []
    res: Result = None
    sca: Result = None
    log = get_logger(__name__)
    async with db() as ses:
        try:
            stmt = (
                Select(
                    MailFilter.filter_id,
                    MailFilter.filter_name,
                    MailFilter.priority,
                    MailFilter.enabled,
                    fc.ifnull(Users.user_name, MailFilter.update_by).label("update_by_name"),
                    MailFilter.update_at,
                )
                .join(
                    Users,
                    MailFilter.update_by == Users.user_id,
                    isouter=True,
                )
                .order_by(MailFilter.priority, MailFilter.filter_name, MailFilter.update_at)
            )
            if lmt > 0:
                stmt = stmt.limit(lmt)
            if off_s > 0:
                stmt = stmt.offset(off_s)

            res = await ses.execute(stmt)
            sca = res.unique()
            for row in sca:
                result.append(MailFilterListItem.model_validate(row))
        except Exception:
            log.exception("メールフィルターの取得に失敗しました。")
            result.clear()
        finally:
            if sca is not None:
                sca.close()
                sca = None
            if res is not None:
                res.close()
                res = None

            await ses.rollback()

    return result


async def get(db: async_sessionmaker[AsyncSession], filter_id: int = -1) -> list[EMailFilter]:
    result: list[EMailFilter] = []
    res: Result = None
    sca: ScalarResult = None
    log = get_logger(__name__)
    async with db() as ses:
        try:
            stmt = (
                Select(MailFilter)
                .options(joinedload(MailFilter.filter_times))
                .options(joinedload(MailFilter.mail_terms))
                .options(joinedload(MailFilter.log_filters))
                .options(joinedload(MailFilter.filter_actions))
                .order_by(MailFilter.priority, MailFilter.filter_name, MailFilter.update_at)
            )
            if filter_id > 0:
                stmt = stmt.where(MailFilter.filter_id == filter_id)

            res = await ses.execute(stmt)
            sca = res.unique().scalars()
            for row in sca:
                result.append(EMailFilter.model_validate(row))
        except Exception:
            log.exception("メールフィルターの取得に失敗しました。")
            result.clear()
        finally:
            if sca is not None:
                sca.close()
                sca = None
            if res is not None:
                res.close()
                res = None
            await ses.rollback()

    return result


async def delete(db: async_sessionmaker[AsyncSession], filter_id: int) -> dict[str, any]:
    result = None
    res: Result = None
    log = get_logger(__name__)
    async with db() as ses:
        try:
            stmt = Select(MailFilter).where(MailFilter.filter_id == filter_id)
            res = await ses.execute(stmt)
            sca = res.unique().scalar_one_or_none()
            if sca is not None:
                await ses.delete(sca)
                await ses.commit()
                await ses.flush()

                result = {"filter_id": filter_id, "status": ResultStatus.delete}
            else:
                await ses.rollback()
                log.warning(f"フィルター ID:{filter_id} のデータが見つかりませんでした。")
        except Exception:
            log.exception("メールフィルターの削除に失敗しました。")
            await ses.rollback()
            result = None
        finally:
            if res is not None:
                res.close()
                res = None

    return result


async def update(db: async_sessionmaker[AsyncSession], ent: EMailFilter) -> dict[str, any]:
    result = None
    res: Result = None
    orm: MailFilter = None

    log = get_logger(__name__)
    async with db() as ses:
        try:
            # DELETE ITEMS
            stmt = Select(MailFilter).where(MailFilter.filter_id == ent.filter_id)
            res = await ses.execute(stmt)
            sca = res.unique().scalar_one_or_none()
            if sca is not None:
                await ses.delete(sca)
            else:
                await ses.rollback()
                log.warning(f"フィルター ID:{ent.filter_id} のデータが見つかりませんでした。")
                return result

            # CREATE ITEMS
            exc = ["update_at"]
            orm = MailFilter.dict_to_model(ent.model_dump(exclude=exc), exclude=exc)
            ses.add(orm)  # 派生テーブルも一括作成
            await ses.commit()
            await ses.flush()
            await ses.refresh(orm)
            result = {"filter_id": orm.filter_id, "status": ResultStatus.update}
        except Exception:
            log.exception("メールフィルターの更新処理 (DEL/INS) に失敗しました。")
            await ses.rollback()
            result = None
        finally:
            if res is not None:
                res.close()
                res = None

    return result


async def get_count(db: async_sessionmaker[AsyncSession], *reject_ids) -> int:
    result: int = -1
    res: Result = None
    async with db() as ses:
        try:
            stmt = Select(func.count()).select_from(MailFilter)
            if len(reject_ids) > 0:
                stmt = stmt.where(MailFilter.filter_id.not_in(reject_ids))
            res = await ses.execute(stmt)
            rows = res.first()
            result = rows[0] if rows is not None else None
        except Exception:
            result = -1
            get_logger(__name__).exception("メールフィルターの登録件数取得に失敗しました。")
        finally:
            if res is not None:
                res.close()
                res = None
            await ses.rollback()

    return result
