from datetime import timedelta

from sqlalchemy import Select, func, Delete, Result, CursorResult
from sqlalchemy.ext.asyncio import async_sessionmaker, AsyncSession

from app import datetime_now_aware, get_logger
from app.models.mail_flow import MailFlow


async def create(db: async_sessionmaker[AsyncSession], **kwd: dict[str, any]) -> bool:
    """データ作成

    メール流量データを作成する

    Parameters
    ----------
    kwd:dict
        filter_id, mail_name のキーを持った dict

    Return
    ------
    bool
        正常終了時、True

    """
    result: bool = False
    async with db() as ses:
        try:
            # 更新日を明示的に設定する
            orm = MailFlow(update_at=datetime_now_aware(), **kwd)
            ses.add(orm)
            await ses.commit()
            await ses.flush()
            orm = None
            result = True
        except Exception:
            get_logger(__name__).exception("メール流量の登録に失敗しました。")
            await ses.rollback()

    return result


async def get_count(db: async_sessionmaker[AsyncSession], filter_id: int, term_minute: int) -> int:
    """データ件数取得

    メール流量の件数を取得する

    Parameters
    ----------
    filter_id:int
        件数を取得するメールフィルターのID
    term_minute:int
        取得対象の時刻(n分前)

    Return
    ------
    int
        件数。エラー時、-1
    """
    result: int = -1
    res: Result = None
    async with db() as ses:
        try:
            curr = datetime_now_aware()
            past = curr - timedelta(minutes=term_minute)
            stmt = (
                Select(func.count())
                .select_from(MailFlow)
                .where(
                    MailFlow.filter_id == filter_id,
                    MailFlow.update_at >= past,
                    MailFlow.update_at <= curr,
                )
            )
            res = await ses.execute(stmt)
            rows = res.first()
            result = rows[0] if rows is not None else None
        except Exception:
            get_logger(__name__).exception("メール流量の件数取得に失敗しました。")
        finally:
            if res is not None:
                if not res.closed:
                    res.close()
                res = None
            await ses.rollback()

    return result


async def delete_at(db: async_sessionmaker[AsyncSession], term_minute: int) -> bool:
    """データ削除

    メール流量データを時間指定で削除する。
    メンテナンス利用を想定。

    Parameters
    ----------
    term_minute:int
        削除対象の時刻(n分前より過去の分)

    Return
    ------
    bool
        正常終了時、True
    """
    result: bool = False
    res: CursorResult = None
    log = get_logger(__name__)
    async with db() as ses:
        try:
            past = datetime_now_aware() - timedelta(minutes=term_minute)
            stmt = Delete(MailFlow).where(MailFlow.update_at <= past)
            res = await ses.execute(stmt)
            await ses.commit()
            await ses.flush()

            log.debug(f"メール流量の削除結果: {res.rowcount}")
            result = True
        except Exception:
            log.exception("メール流量の削除に失敗しました。")
            await ses.rollback()
        finally:
            if res is not None:
                res.close()
                res = None

    return result
