from sqlalchemy import Select, Result, ScalarResult
from sqlalchemy.ext.asyncio import async_sessionmaker, AsyncSession

# from memory_profiler import profile

from app import get_logger
from app.enum import ResultStatus, MailProtocol
from app.models.mail_server import MailServer as MMailServer
from app.entity.mail_server import MailServer as EMailServer
from app.service.configure import SayTorConfig


async def create(db: async_sessionmaker[AsyncSession], ent: EMailServer) -> dict[str, any]:
    result: dict = None
    orm: MMailServer = None
    async with db() as ses:
        try:
            orm = MMailServer(**ent.model_dump(exclude=["password", "update_at"]))
            orm.password = SayTorConfig().encrypt(ent.password)
            ses.add(orm)
            await ses.commit()
            result = {"host": ent.host, "port": ent.port, "status": ResultStatus.create}
        except Exception:
            get_logger(__name__).exception("メールサーバー設定の登録に失敗しました。")
            await ses.rollback()
            result = None
    orm = None

    return result


async def get(db: async_sessionmaker[AsyncSession], host: str, port: int) -> EMailServer:
    result = None
    log = get_logger(__name__)
    res: Result = None

    async with db() as ses:
        try:
            stmt = Select(MMailServer).where(MMailServer.host == host, MMailServer.port == port)
            res = await ses.execute(stmt)
            sca = res.scalar_one_or_none()
            if sca is not None:
                result = EMailServer.model_validate(sca)
                result.password = SayTorConfig().decrypt(result.password)
            else:
                log.warning(f"メールサーバー情報 [{host}:{port}] が見つかりませんでした。")
        except Exception:
            log.exception("メールサーバー情報の取得に失敗しました。")
        finally:
            if res is not None:
                res.close()
                res = None
            await ses.rollback()

    return result


# @profile(precision=4)
async def get_as_protocol(db: async_sessionmaker[AsyncSession], proto: MailProtocol) -> list[EMailServer]:
    result: list[EMailServer] = []
    res: Result = None
    sca: ScalarResult = None

    cfg = SayTorConfig()
    async with db() as ses:
        try:
            stmt = Select(MMailServer).where(MMailServer.protocol == proto.name)
            res = await ses.execute(stmt)
            sca = res.scalars()
            for row in sca:
                m = EMailServer.model_validate(row)
                m.password = cfg.decrypt(m.password)
                result.append(m)
        except Exception:
            result = []
            get_logger(__name__).exception("メールサーバー情報の一括取得に失敗しました。")
        finally:
            if sca is not None:
                sca.close()
                sca = None
            if res is not None:
                res.close()
                res = None

            await ses.rollback()

    return result


async def update(db: async_sessionmaker[AsyncSession], ent: EMailServer) -> dict[str, any]:
    result = None
    res: Result = None
    log = get_logger(__name__)
    async with db() as ses:
        try:
            stmt = Select(MMailServer).where(MMailServer.host == ent.host, MMailServer.port == ent.port)
            res = await ses.execute(stmt)
            sca = res.scalar_one_or_none()
            if sca is not None:
                sca.update(**ent.model_dump(exclude=["password"]))
                sca.password = SayTorConfig().encrypt(ent.password)
                await ses.commit()

                result = {"host": ent.host, "port": ent.port, "status": ResultStatus.update}
            else:
                await ses.rollback()
                log.warning(f"メールサーバー情報 [{ent.host}:{ent.port}] が見つかりませんでした。")
        except Exception:
            log.exception("メールサーバー情報の更新に失敗しました。")
            await ses.rollback()
            result = None
        finally:
            if res is not None:
                res.close()
                res = None

    return result


async def delete(db: async_sessionmaker[AsyncSession], host: str, port: int) -> dict[str, any]:
    result = None
    res: Result = None
    log = get_logger(__name__)
    async with db() as ses:
        try:
            stmt = Select(MMailServer).where(MMailServer.host == host, MMailServer.port == port)
            res = await ses.execute(stmt)
            sca = res.scalar_one_or_none()
            if sca is not None:
                await ses.delete(sca)
                await ses.commit()
                await ses.flush()

                result = {"host": host, "port": port, "protocol": sca.protocol, "status": ResultStatus.delete}
            else:
                await ses.rollback()
                log.warning(f"メールサーバー情報 [{host}:{port}] が見つかりませんでした。")
        except Exception:
            log.exception("メールサーバー情報の削除に失敗しました。")
            await ses.rollback()
            result = None
        finally:
            if res is not None:
                res.close()
                res = None

    return result
