from sqlalchemy import Select, Result
from sqlalchemy.ext.asyncio import async_sessionmaker, AsyncSession
from sqlalchemy.exc import IntegrityError

from app import get_logger
from app.enum import ResultStatus
from app.models.users import Users as MUser
from app.entity.users import ChangePassword, Users as EUser


async def create(db: async_sessionmaker[AsyncSession], ent: EUser) -> dict[str, str]:
    result = None
    orm: MUser = None
    log = get_logger(__name__)
    async with db() as ses:
        try:
            # パスワードのハッシュ化
            ent.hash_pass()

            orm = MUser(**ent.model_dump(exclude=["update_at"]))
            ses.add(orm)
            await ses.commit()
            await ses.flush()
            await ses.refresh(orm)

            result = {"user": orm.user_id, "status": ResultStatus.create}
        except IntegrityError:
            log.error(f"ユーザー ID: {ent.user_id} が重複しました。有効/無効:{ent.enabled}")
            result = None
            await ses.rollback()
        except Exception:
            log.exception("ユーザー情報の登録に失敗しました。")
            result = None
            await ses.rollback()

    orm = None
    return result


async def get(db: async_sessionmaker[AsyncSession], user_id: str) -> EUser:
    result = None
    res: Result = None
    async with db() as ses:
        try:
            stmt = Select(MUser).where(MUser.user_id == user_id)
            res = await ses.execute(stmt)
            sca = res.scalar_one_or_none()
            if sca is not None:
                result = EUser.model_validate(sca)
        except Exception:
            get_logger(__name__).exception("ユーザー情報の取得に失敗しました。")
        finally:
            if res is not None:
                res.close()
                res = None
            await ses.rollback()

    return result


async def update(db: async_sessionmaker[AsyncSession], ent: EUser) -> dict[str, str]:
    result = None
    res: Result = None
    log = get_logger(__name__)
    async with db() as ses:
        try:
            stmt = Select(MUser).where(MUser.user_id == ent.user_id)
            res = await ses.execute(stmt)
            sca = res.scalar_one_or_none()
            if sca is not None:
                sca.update(**ent.model_dump())
                await ses.commit()
                await ses.flush()
                await ses.refresh(sca)

                result = {"user": sca.user_id, "status": ResultStatus.update}
            else:
                await ses.rollback()
                log.warning(f"ユーザー ID: {ent.user_id} のユーザー情報が見つかりませんでした。")
        except IntegrityError:
            log.error("ユーザー ID: {ent.user_id} が重複しました。有効/無効:{ent.enabled}")
            result = None
            await ses.rollback()
        except Exception:
            log.exception("ユーザー情報の更新に失敗しました。")
            await ses.rollback()
            result = None
        finally:
            if res is not None:
                res.close()
                res = None

    return result


async def change_password(db: async_sessionmaker[AsyncSession], ent: ChangePassword) -> dict[str, str]:
    result = None
    res: Result = None
    log = get_logger(__name__)
    async with db() as ses:
        try:
            stmt = Select(MUser).where(MUser.user_id == ent.user_id)
            res = await ses.execute(stmt)
            sca = res.scalar_one_or_none()
            if sca is not None:
                if sca.secret_key.casefold() != ent.old_key:
                    log.error(f"ユーザー ID: {ent.user_id} のパスワードが一致しません。")
                else:
                    sca.secret_key = ent.secret_key
                    sca.update_by = ent.user_id
                    await ses.commit()
                    await ses.flush()

                    result = {"user": ent.user_id, "status": ResultStatus.update}

            if result is None:
                await ses.rollback()
                log.warning(f"ユーザーID: {ent.user_id} は変更されませんでした。")
        except Exception:
            log.exception("パスワードの変更に失敗しました。")
            await ses.rollback()
            result = None
        finally:
            if res is not None:
                res.close()
                res = None

    return result
