import json
from asyncio import current_task
from contextlib import asynccontextmanager

from sqlalchemy.engine import make_url
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.asyncio import async_sessionmaker, async_scoped_session

from app import get_logger
from app.service.configure import SayTorConfig
from app.util import STORE_ROOT
from app.util.env_config import set_env

# initialize
DEFAULT_DB_STR = "sqlite+aiosqlite:///{store_dir}/dashboard.db"
Engine = None
DbSession = None


def get_connection_string() -> str:
    cfg = SayTorConfig()

    # 接続設定の取得
    words: str = cfg.get_config("database")
    if len(words) <= 0:
        words = DEFAULT_DB_STR.format(store_dir=STORE_ROOT).replace("\\", "/")

    # 接続設定の再保存(この時、非暗号化のデータは暗号化される)
    set_env("database", cfg.encrypt(words))

    get_logger(__name__).debug(f"データベース接続用 URI: {words}")
    return make_url(words)


def connect() -> bool:
    result: bool = False
    global Engine
    global DbSession

    try:
        db_url = get_connection_string()
        db_opt = json.loads(SayTorConfig().get_config("conn_args", "{}"))

        ## 接続開始 #######################################################################
        Engine = create_async_engine(db_url, **db_opt)
        DbSession = async_scoped_session(
            async_sessionmaker(
                autocommit=False,
                bind=Engine,
                class_=AsyncSession,
                future=True,
            ),
            scopefunc=current_task,
        )
        result = True
    except Exception:
        get_logger(__name__).exception("データベースへの接続に失敗しました。")

    return result


@asynccontextmanager
async def get_ses_maker():
    try:
        yield DbSession
    finally:
        if DbSession:
            await DbSession.remove()
