
from datetime import datetime

from pydantic import ValidationError, BaseModel, ConfigDict, field_serializer
from pydantic.alias_generators import to_camel

from app import tzinfo

class CommonModel(BaseModel):
    """Modelクラス共通定義

    管理サーバーからのパクリ。

    """

    model_config = ConfigDict(
        populate_by_name=True,          # 別名定義ON
        from_attributes=True,           # ormモードON
        alias_generator=to_camel,       # スネークケースをローワーキャメルケースとして別名定義させる
        str_strip_whitespace=True,      # ホワイトスペースのトリミング
        ser_json_timedelta="iso8601",   # シリアライズは基本 ISO8601 形式
    )

    @field_serializer('update_at', check_fields=False)
    def serialize_update_at(self, dt: datetime, _info):
        """jsonエンコード時の設定

        datetimeをISO 8601+タイムゾーン表記にする

        Parameters
        ----------
        dt: datetime
            変換対象の日時
        """
        return dt.replace(microsecond=0, tzinfo=tzinfo()).isoformat()

def get_validate_errors(errs:ValidationError) -> dict[str, list[str]]:
    lst:dict[list[str]] = {}
    for e in errs.errors():
        typ:str = e["type"]
        if typ.endswith("_too_short"):
            msg = "値が入力されていません。" if typ.startswith("string") else "値が小さすぎます。"
        elif typ.endswith("_too_large"):
            msg = "入力文字数が多すぎます。" if typ.startswith("string") else "値が大きすぎます。"
        elif typ.endswith("_type"):
            msg = "入力形式が不正です。"
        elif typ.startswith("missing"):
            msg = "値が入力されていません。"
        else:
            msg = typ

        for f in e["loc"]:
            if f not in lst.keys():
                lst[f] = []
            lst[f].append(msg)
    
    return lst
