import os
import re
import shutil
from datetime import time, datetime
from typing import Optional

from pydantic import Field, constr, field_serializer, conint, computed_field

# from memory_profiler import profile

from app import datetime_now_aware, tzinfo, SYSTEM_USER_ID, get_logger
from app.enum import TermType, TimeType, FieldType, TermJoin, ActionType
from app.entity import CommonModel
from app.service.attach_file import AttachFiles
from app.util.log_file import LogFile
from app.service.recv_data import BaseReadMail
from app.service.send_eml import SendMailMsg
from app.service.calendar import SystemCalendar
from app.util.file_util import get_temp_dir, peaceably_delete, peaceably_rd, peaceably_clear


class FilterAction(CommonModel):
    filter_id: int = Field(0, description="フィルターID")
    act_type: conint(ge=min(list(map(int, ActionType))), le=max(list(map(int, ActionType))))
    act_data: constr(strip_whitespace=False, min_length=1, max_length=4096)

    update_by: str = Field(SYSTEM_USER_ID, description="更新者ID")
    update_at: datetime = Field(
        description="更新日時",
        default_factory=lambda: datetime_now_aware(),
    )


class FilterTimes(CommonModel):
    filter_id: int = Field(0, description="フィルターID")
    term_no: conint(ge=1, le=99)
    time_type: conint(ge=min(list(map(int, TimeType))), le=max(list(map(int, TimeType))))
    time_from: time = Field(time(0, 0, 0), description="開始時刻")
    time_to: time = Field(time(23, 59, 59), description="終了時刻")

    update_by: str = Field(SYSTEM_USER_ID, description="更新者ID")
    update_at: datetime = Field(
        description="更新日時",
        default_factory=lambda: datetime_now_aware(),
    )

    @field_serializer("time_from", "time_to", check_fields=False)
    def serial_from_time(self, dt: time, _info):
        return dt.replace(microsecond=0, tzinfo=tzinfo()).isoformat()


class MailTerms(CommonModel):
    filter_id: int = Field(0, description="フィルターID")
    term_no: conint(ge=1, le=99)
    field_type: conint(ge=min(list(map(int, FieldType))), le=max(list(map(int, FieldType))))
    keywords: constr(strip_whitespace=True, min_length=1, max_length=64)
    term_type: conint(ge=min(list(map(int, TermType))), le=max(list(map(int, TermType))))

    update_by: str = Field(SYSTEM_USER_ID, description="更新者ID")
    update_at: datetime = Field(
        description="更新日時",
        default_factory=lambda: datetime_now_aware(),
    )


class LogTerms(CommonModel):
    filter_id: int = Field(0, description="フィルターID")
    term_no: conint(ge=1, le=99)
    keywords: constr(min_length=1, max_length=64)
    term_type: conint(ge=min(list(map(int, TermType))), le=max(list(map(int, TermType))))

    update_by: str = Field(SYSTEM_USER_ID, description="更新者ID")
    update_at: datetime = Field(
        description="更新日時",
        default_factory=lambda: datetime_now_aware(),
    )


def hit(line: str, term) -> bool:
    """条件一致判定

    文字列が条件と一致するかどうか判定する

    Parameters
    ----------
    line:str
        判定対象の文字列
    term
        MailTerms, LogTerms インスタンス

    Return
    ------
    bool
        条件に合致した場合、True
    """
    result: bool = False
    if term.term_type == TermType.included:
        result = term.keywords.casefold() in line.casefold()
    elif term.term_type == TermType.not_included:
        result = term.keywords.casefold() not in line.casefold()
    elif term.term_type == TermType.match_text:
        result = line.casefold() == term.keywords.casefold()
    elif term.term_type == TermType.not_match_text:
        result = line.casefold() != term.keywords.casefold()
    elif term.term_type == TermType.match_regexp:
        result = re.search(term.keywords, line, re.I) is not None
    elif term.term_type == TermType.not_match_regexp:
        result = re.search(term.keywords, line, re.I) is None

    return result


class MailFilter(CommonModel):
    filter_id: int = Field(0, description="フィルターID")
    filter_name: constr(strip_whitespace=True, min_length=1, max_length=32)
    priority: conint(ge=1, le=9999)
    enabled: bool = Field(True, description="フィルター有効")
    mail_term_join: conint(ge=min(list(map(int, TermJoin))), le=max(list(map(int, TermJoin))))
    enable_flow: bool = Field(False, description="流量有効")
    flow_interval: Optional[int] = Field(5, description="間隔（分）")
    flow_count: Optional[int] = Field(None, description="流量数")
    flow_term: Optional[int] = Field(None, description="流量条件フラグ")
    log_term_join: conint(ge=min(list(map(int, TermJoin))), le=max(list(map(int, TermJoin))))
    send_empty_log: bool = Field(False, description="ログ無転送許可")

    filter_times: list[FilterTimes] = Field([], description="起動指定時間リスト")
    mail_terms: list[MailTerms] = Field([], description="メール条件リスト")
    log_filters: list[LogTerms] = Field([], description="ログフィルター条件リスト")
    filter_actions: list[FilterAction] = Field([], description="フィルターアクション")

    update_by: str = Field(SYSTEM_USER_ID, description="更新者ID")
    update_at: datetime = Field(
        description="更新日時",
        default_factory=lambda: datetime_now_aware(),
    )

    def penetrate_filter_id(self):
        """フィルターID浸透

        関連エンティティに自身のフィルターIDを設定する
        """
        if self.filter_times:
            for fs in self.filter_times:
                fs.filter_id = self.filter_id
        if self.mail_terms:
            for fs in self.mail_terms:
                fs.filter_id = self.filter_id
        if self.log_filters:
            for fs in self.log_filters:
                fs.filter_id = self.filter_id
        if self.filter_actions:
            for fs in self.filter_actions:
                fs.filter_id = self.filter_id

    def penetrate_update_by(self, uid: str):
        """更新者ID浸透

        Parameters
        ----------
        uid:str
            更新者ID
        """
        self.update_by = uid
        if self.filter_times:
            for fs in self.filter_times:
                fs.update_by = uid
        if self.mail_terms:
            for fs in self.mail_terms:
                fs.update_by = uid
        if self.log_filters:
            for fs in self.log_filters:
                fs.update_by = uid
        if self.filter_actions:
            for fs in self.filter_actions:
                fs.update_by = uid

    def is_executable(self) -> bool:
        """起動判定

        当該フィルターが起動可能か確認する

        Return
        ------
        bool
            起動可能な場合、True
        """
        log = get_logger(__name__)
        if not self.enabled:
            log.detail(f"\tフィルター ID: {self.filter_id} は無効。")
            return False
        elif len(self.filter_times) <= 0:
            log.detail(f"\tフィルター ID: {self.filter_id} の時刻設定はありません。有効です。")
            return True

        # 日時取得
        curr_dt = datetime_now_aware()
        curr_tm = curr_dt.time()
        calendar = SystemCalendar()
        le_time = lambda x, y: x.replace(second=0, microsecond=0) <= y.replace(second=0, microsecond=0)

        result: bool = False
        for t in self.filter_times:
            # 条件判定
            if not t.time_type == TimeType.everyday:
                if t.time_type == TimeType.weekday and curr_dt.weekday() >= 5:
                    # 平日指定で週末
                    continue
                elif t.time_type == TimeType.weekend and curr_dt.weekday() < 5:
                    # 週末指定で平日
                    continue
                elif t.time_type == TimeType.workday and calendar.is_holiday():
                    #  営業日指定で休日
                    continue
                elif t.time_type == TimeType.holiday and calendar.is_workday():
                    # 休日指定で営業日
                    continue

            if le_time(t.time_from, curr_tm) and le_time(curr_tm, t.time_to):
                # 時刻範囲に現在時刻があれば OK
                result = True
                break

        log.detail(f"\tフィルター ID: {self.filter_id} の時刻設定確認結果={result}")
        return result

    def valid(self, data: BaseReadMail) -> bool:
        """フィルター対象判定

        指定されたデータがこのフィルターと合致するか判定する

        Parameters
        ----------
        data:ReadMailFile
            判定対象のデータ

        Return
        ------
        bool
            対象の場合、True
        """
        log = get_logger(__name__)
        result = False

        hit_count: int = 0

        for mt in self.mail_terms:
            # データ取得
            txt: str = ""
            if mt.field_type == FieldType.field_from:
                txt = data.get_addr_from()
                log.detail(f"\t対象項目 'from' の値: '{txt}'")
            elif mt.field_type == FieldType.field_to:
                txt = data.get_addr_to()
                log.detail(f"\t対象項目 'to' の値: '{txt}'")
            elif mt.field_type == FieldType.field_cc:
                txt = data.get_addr_cc()
                log.detail(f"\t対象項目 'cc' の値: '{txt}'")
            elif mt.field_type == FieldType.field_subject:
                txt = data.get_subject()
                log.detail(f"\t対象項目 'subject' の値: '{txt}'")
            elif mt.field_type == FieldType.field_body:
                txt = data.get_body()
                log.detail(f"\t対象項目 'body' の値: '{txt}'")

            # 条件判定
            if hit(txt, mt):
                hit_count += 1
                log.detail(f"\t条件に合致。ヒットした条件数={hit_count}")
            else:
                log.detail(f"\t条件は一致せず。ヒットした条件数={hit_count}")

        if self.mail_term_join == TermJoin.join_and:
            # 全ての条件に合致
            result = hit_count == len(self.mail_terms)
        else:
            # どれか一つに合致
            result = hit_count > 0

        log.detail(f"\tフィルター ID: {self.filter_id} フィルター合致判定結果= {result}")
        return result

    # @profile(precision=4)
    def exec_filter(self, file_path: str) -> str:
        """ファイルフィルター処理

        ファイルに対してフィルター処理を実施する。

        Parameters
        ----------
        file_path:str
            対象ファイルのパス

        Return
        ------
        str
            フィルター後のファイルパス。空っぽになった場合は空文字
        """
        console = get_logger(__name__)
        console.debug(f"ログファイル {os.path.basename(file_path)} の確認開始。")

        # ログファイルの処理
        log = LogFile(file_path)
        if log is None:
            # バイナリと判断したので何も処理しない
            return file_path

        while not log.is_eof():
            txt = log.get_line()
            console.detail(f"\tLINE 【{txt}】")

            hit_count = 0
            for lt in self.log_filters:
                # 条件判定
                if hit(txt, lt):
                    hit_count += 1
            console.detail(f"\tHIT 【{hit_count}/{len(self.log_filters)}】")

            added: bool = True
            if self.log_term_join == TermJoin.join_and:
                # 全ての条件に合致
                added = not (hit_count == len(self.log_filters))
            else:
                # どれか一つに合致
                added = not hit_count > 0

            if added:
                console.detail("\t\t当該行は保持")
                log.keep()
            else:
                console.detail("\t\t当該行は削除")

        # 処理結果を保存
        log.save()

        return log.path if os.path.exists(log.path) else ""

    # @profile(precision=4)
    def attach_filter(self, att: AttachFiles) -> AttachFiles:
        """添付ファイルフィルター処理

        添付ファイルに対するフィルター処理を制御する

        Parameters
        ----------
        att:AttachFiles
            添付ファイル管理クラス

        Return
        ------
        AttachFile
            フィルター処理した結果の添付ファイルクラス
        """
        console = get_logger(__name__)

        # 添付内のログファイルを処理する
        for f in att.pop_log():
            # ログファイルの処理
            new_name = self.exec_filter(f)
            if len(new_name) > 0:
                att.modified(new_name)

        # 添付内の圧縮ファイルを処理する
        arc_temp_dir: str = get_temp_dir()
        for arc in att.pop_arch():
            if arc is None:
                continue

            console.debug(f"圧縮添付ファイル {os.path.basename(arc.path)} の確認開始。")

            # 圧縮ファイルを伸張する
            if not arc.store(arc_temp_dir):
                # 失敗したら、再度管理下に入れ直して次のファイルを処理する
                console.error("圧縮添付ファイルの伸張でエラーが発生しました。当該ファイルはスキップされます。")
                att.modified(arc.path)
                continue
            else:
                console.debug(f"extract result:{str(arc.files)}")

            # 添付用の管理クラスで展開したファイルを処理させる
            a2 = self.attach_filter(AttachFiles(arc.files))
            if a2.file_count() > 0:
                # 再圧縮(ZIP)
                if not a2.to_zip(arc.path):
                    console.error("処理後の添付ファイルの圧縮でエラーが発生しました。当該ファイルはスキップされます。")
                else:
                    console.debug(f"圧縮添付ファイル {arc.path} の処理が終了しました。")
                att.modified(arc.path)
            else:
                # ファイルが空
                console.warning("処理した結果、圧縮添付ファイル内が空になりました。")
                peaceably_delete(arc.path)  # 念のため
            peaceably_rd(arc_temp_dir, True)

        # 最終的には全削除
        peaceably_rd(arc_temp_dir)

        # 添付の結果を返す
        return att

    def check_mail_attach(self, data: BaseReadMail) -> SendMailMsg:
        """添付ファイルフィルター処理

        添付ファイルをフィルター処理する。

        Parameters
        ----------
        data:ReadMailFile
            判定対象のメールデータ

        Return
        ------
        SendMailMsg
            送信用メール。送信しない場合、None
        """
        console = get_logger(__name__)

        if len(self.log_filters) <= 0:
            # ログフィルターは無いので、メールは返却しない
            console.debug("ログフィルターは未設定のため、メール情報は返却されません。")
            return None

        # 全添付を保存
        lst: list[str] = data.store_all_attach()

        # 添付が無い場合
        if len(lst) <= 0:
            console.debug("添付ファイルがありません。")
            # ログ無転送を許可していれば、メールを作成する
            if self.send_empty_log:
                console.debug("\tが、ログ無し転送が許可されているので、転送メール情報を返却します。")
                return SendMailMsg(
                    subject=data.get_subject(),
                    body=data.get_body(),
                )

            # 添付無しなので、送信しない
            return None

        # フィルター処理を実施
        att = self.attach_filter(AttachFiles(lst))

        # ファイルのフィルター結果
        ret_mail = None

        # 返却用メール生成
        if att.file_count() > 0:
            ret_mail = SendMailMsg(subject=data.get_subject(), body=data.get_body(), attach_file=att.all_files())
        elif self.send_empty_log:
            console.debug("添付ファイルはありませんが、ログ無し転送が許可されているので転送メール情報を返却します。")
            ret_mail = SendMailMsg(
                subject=data.get_subject(),
                body=data.get_body(),
            )
        else:
            console.debug("転送メールの情報はありません。")
            peaceably_clear(lst)

        return ret_mail


def create_empty_filter() -> MailFilter:
    """新規用ブランクエンティティ作成"""
    result: MailFilter = MailFilter(
        filter_name="新規フィルター",
        priority=500,
        enabled=True,
        mail_term_join=TermJoin.join_and,
        enable_flow=False,
        log_term_join=TermJoin.join_or,
        send_empty_log=False,
        filter_times=[],
        mail_terms=[],
        log_filters=[],
        filter_actions=[],
    )

    return result


class MailFilterListItem(CommonModel):
    """フィルター一覧用項目エンティティ"""

    filter_id: int = Field(0, description="フィルターID")
    filter_name: str = Field("", description="フィルター名")
    priority: int = Field(500, description="優先度")
    enabled: bool = Field(True, description="フィルター有効")
    update_by_name: Optional[str] = Field("", description="更新者")
    update_at: datetime = Field(description="更新日")

    @computed_field
    @property
    def rollover_at(self) -> datetime:
        if self.update_at:
            return self.update_at.replace(microsecond=0)
        return None
