
from datetime import datetime, time

from sqlalchemy import Enum, event
from sqlalchemy import Column, String, Integer, Boolean, BigInteger, Identity, Index, ForeignKey, Time
from sqlalchemy.orm import relationship

from app.enum import TermJoin, FlowTerm, TimeType, FieldType,TermType, ActionType
from app.models import Base, UpdatesMixin, modified_at_hook

class FilterAction(Base, UpdatesMixin):
    """フィルターアクション
    """
    __tablename__ = "actions"
    
    filter_id = Column(BigInteger, ForeignKey("mail_filter.filter_id"), nullable=False, primary_key=True, index=True)
    act_type = Column(Enum(ActionType), nullable=False, primary_key=True, default=ActionType.action_mail.name)
    act_data = Column(String(4096), nullable=False)

    UPDATE_FIELDS = [
        "act_data",
        "update_by",
    ]

class FilterTimes(Base, UpdatesMixin):
    """起動時刻
    """
    __tablename__ = "filter_times"
    
    filter_id = Column(BigInteger, ForeignKey("mail_filter.filter_id"), nullable=False, primary_key=True, index=True)
    term_no = Column(Integer, nullable=False, primary_key=True)
    time_type = Column(Enum(TimeType), nullable=False, default=TimeType.everyday.name)
    time_from = Column(Time, nullable=False)
    time_to = Column(Time, nullable=False)

    UPDATE_FIELDS = [
        "time_type",
        "time_from",
        "time_to",
        "update_by",
    ]

class MailTerms(Base, UpdatesMixin):
    """メール条件
    """
    __tablename__ = "mail_terms"
    
    filter_id = Column(BigInteger, ForeignKey("mail_filter.filter_id"), nullable=False, primary_key=True, index=True)
    term_no = Column(Integer, nullable=False, primary_key=True)
    field_type = Column(Enum(FieldType), nullable=False, default=FieldType.field_from.name)
    keywords = Column(String(64), nullable=False)
    term_type = Column(Enum(TermType), nullable=False, default=TermType.included.name)

    UPDATE_FIELDS = [
        "field_type",
        "keywords",
        "term_type",
        "update_by",
    ]

class LogTerms(Base, UpdatesMixin):
    """ログフィルター条件
    """
    __tablename__ = "log_terms"

    filter_id = Column(BigInteger, ForeignKey("mail_filter.filter_id"), nullable=False, primary_key=True, index=True)
    term_no = Column(Integer, nullable=False, primary_key=True)
    keywords = Column(String(64), nullable=False)
    term_type = Column(Enum(TermType), nullable=False, default=TermType.included.name)

    UPDATE_FIELDS = [
        "term_no",
        "keywords",
        "term_type",
        "update_by",
    ]

class MailFilter(Base, UpdatesMixin):
    """メールフィルター
    """
    __tablename__ = "mail_filter"

    # 末尾の ,) がミソ
    __table_args__ = (Index("idx_mail_flt_1",  "priority", "filter_name", "update_at"),)
    
    filter_id = Column(
        BigInteger().with_variant(Integer, "sqlite"), 
        Identity(start=1, cycle=False), 
        nullable=False, 
        primary_key=True, 
        autoincrement="auto")
    filter_name = Column(String(32), nullable=False)
    priority = Column(Integer, nullable=False, default=500)
    enabled = Column(Boolean, nullable=False, default=True)
    mail_term_join = Column(Enum(TermJoin), nullable=False, default=TermJoin.join_and.name)
    enable_flow = Column(Boolean, nullable=False, default=False)
    flow_interval = Column(Integer, nullable=True, default=5)
    flow_count = Column(Integer, nullable=True)
    flow_term = Column(Enum(FlowTerm), nullable=True)
    log_term_join = Column(Enum(TermJoin), nullable=False, default=TermJoin.join_and.name)
    send_empty_log = Column(Boolean, nullable=False, default=False)

    # DEL/INS を使うので、Updateは無
    UPDATE_FIELDS = []

    # 起動指定時間リスト
    filter_times = relationship(
        "FilterTimes",                  # 参照先のModelクラス名を指定
        cascade="all, delete-orphan",   # 外部参照設定(delete-orphan: 1:n のテーブルでカスケード削除)
    )

    # メール条件リスト
    mail_terms = relationship(
        "MailTerms",
        cascade="all, delete-orphan",
    )

    # ログフィルターリスト
    log_filters = relationship(
        "LogTerms",
        cascade="all, delete-orphan",
    )

    # ログアクションリスト
    filter_actions = relationship(
        "FilterAction",
        cascade="all, delete-orphan",
    )

    @classmethod
    def dict_to_model(cls, data:dict[str,any], exclude:list[str]=[]):
        # リレーション以外を設定する
        extract = lambda n, d: datetime.fromisoformat(d[n]) if n == "update_at" else d[n]
        self = cls()
        filter_id = data["filter_id"] if "filter_id" in data else None
        if not "filter_id" in exclude:
            self.filter_id = filter_id

        for k in [
                "filter_name",
                "priority",
                "enabled",
                "mail_term_join",
                "enable_flow",
                "flow_interval",
                "flow_count",
                "flow_term",
                "log_term_join",
                "send_empty_log",
                "update_by",
                "update_at"]:
            if k in data:
                if k not in exclude:
                    setattr(self, k, extract(k, data))

        # リレーション：起動時刻
        self.filter_times = []
        if "filter_times" in data:
            for d in data["filter_times"]:
                m = FilterTimes()
                for k in d.keys():
                    if k not in exclude:
                        if k in ["time_from", "time_to"]:
                            val = time.fromisoformat(d[k]) if isinstance(d[k], str) else d[k]
                            setattr(m, k, val)
                        else:
                            setattr(m, k, extract(k, d))
                if "filter_id" not in exclude:
                    m.filter_id = filter_id
                self.filter_times.append(m)

        # リレーション：メールフィルター条件
        self.mail_terms = []
        if "mail_terms" in data:
            for d in data["mail_terms"]:
                m = MailTerms()
                for k in d.keys():
                    if k not in exclude:
                        setattr(m, k, extract(k, d))
                if "filter_id" not in exclude:
                    m.filter_id = filter_id
                self.mail_terms.append(m)

        # リレーション：ログフィルター条件
        self.log_filters = []
        if "log_filters" in data:
            for d in data["log_filters"]:
                m = LogTerms()
                for k in d.keys():
                    if k not in exclude:
                        setattr(m, k, extract(k, d))
                if "filter_id" not in exclude:
                    m.filter_id = filter_id
                self.log_filters.append(m)

        # リレーション：フィルターアクション
        self.filter_actions = []
        if "filter_actions" in data:
            for d in data["filter_actions"]:
                m = FilterAction()
                for k in d.keys():
                    if k not in exclude:
                        setattr(m, k, extract(k, d))
                if "filter_id" not in exclude:
                    m.filter_id = filter_id
                self.filter_actions.append(m)

        return self

# event
event.listen(MailFilter, "before_insert", modified_at_hook)
event.listen(MailFilter, "before_update", modified_at_hook)
