import json
from base64 import urlsafe_b64encode
from http import HTTPStatus
from urllib.parse import urlparse

import tornado.web
from tornado.escape import json_decode, json_encode, utf8
from tornado.util import unicode_type

from app import get_logger, COOKIE_NAME, EXPIRE_DAY, SYSTEM_USER_ID, APP_NAME, random_string
from app.db import get_ses_maker
from app.crud.users import get as user_get
from app.entity.auth_info import AuthInfo
from app.entity.sys_err_res import SysErrorResponse
from app.service.configure import SayTorConfig
from app.service.license import is_trial, get_remain_day
from app.util import convert_password
from app.util.b64_util import get_bytes

# API キー認証用
API_KEYS = [
    "gAAAAABlAUPYGdUHbMfwBIyse36Yjh8P9OnY9s4qS0intt8TfFqBhfGhNyyVcLOif1OQKQbZaFFsvYlDkKlHG6rL0LDxpkN5ibdAvk4JsXh7yGKsEVgQQsM7Gj_rxpKLZ2UKFNHn8h7X"
]

BROWSER_IDENTITY = ("msie", "trident", "edg", "edge", "chrome", "safari", "firefox")

ITEM_NOT_FOUND_ERROR = tornado.web.HTTPError(HTTPStatus.NOT_FOUND, "404 Page not found.")

ABEND_SAVE_ERROR = tornado.web.HTTPError(
    HTTPStatus.INTERNAL_SERVER_ERROR, "500 Internal server error, failed preserve DB"
)

SERVER_RUNTIME_ERROR = tornado.web.HTTPError(HTTPStatus.SERVICE_UNAVAILABLE, "503 Service unavailable.")

GET_UID = lambda x: x[0] if x is not None else SYSTEM_USER_ID


def NO_PATH_PARAMS(*args, **kwargs):
    if len(args) > 0 or len(kwargs) > 0:
        raise ITEM_NOT_FOUND_ERROR


class AlertFilterHeaderTransform(tornado.web.OutputTransform):
    """レスポンスヘッダーフック

    アラートフィルター用のレスポンスヘッダーフック

    """
    def __init__(self, request):
        # リクエストパスを取得する
        self.path:str = urlparse(request.uri).path

    def transform_first_chunk(self, status_code, headers, chunk, finishing):
        """レスポンスヘッダー操作

        セキュリティ上、設定するレスポンスヘッダーを事前設定

        """
        # サーバーヘッダー削除
        headers.pop("Server")

        # JavaScript の ContentType を強制するための対処
        if self.path.endswith(".js"):
            headers["Content-Type"] = "text/javascript"

        # 設定済みのヘッダー script-src を取得する
        already_head: str = ""
        if "Content-Security-Policy" in headers.keys():
            already_head = headers.pop("Content-Security-Policy")
        else:
            already_head = "script-src 'self'"

        # セキュリティヘッダー追加
        headers["Content-Security-Policy"] = "; ".join(
            [
                "frame-ancestors 'self'",
                "object-src 'none'",
                "style-src 'self' https://fonts.googleapis.com/ https://fonts.gstatic.com/",
                "img-src 'self'",
                "connect-src 'self'",
                "form-action 'self'",
                "font-src https://fonts.googleapis.com/ https://fonts.gstatic.com/ 'self'",
                "frame-src 'none'",
                "media-src 'none'",
                "manifest-src 'self'",
                already_head,
            ]
        )

        # Content-Type-Option 追加
        headers["X-Content-Type-Options"] = "nosniff"

        # キャッシュコントロール(キャッシュNG)
        headers["Cache-Control"] = "no-store"
        headers["Pragma"] = "no-cache"

        # SSL 利用時のみ
        if SayTorConfig().use_ssl():
            # 1 時間有効
            headers["Strict-Transport-Security"] = "max-age=3600; preload"

        return status_code, headers, chunk


async def api_auth(username, password) -> tuple[str]:
    """認証本体

    認証情報を確認する

    Parameters
    ----------
    username:str
        ユーザー名、または API キー
    password:str
        暗号化されたパスワード
    """
    log = get_logger(__name__)
    if username in API_KEYS:
        log.info("API キー認証が行われます。")
        return (SYSTEM_USER_ID, APP_NAME)

    # ユーザー情報を取得する
    async with get_ses_maker() as db:
        usr = await user_get(db, username)

    if usr is None:
        log.warning(f"ユーザーID: {username} の情報がありません。")
        return None

    # パスワードを取得する
    pass_ = convert_password(password)

    # パスワードの同一確認
    if usr.secret_key == pass_:
        log.info(f"ユーザーID: {username} は認証されました。")
        return (usr.user_id, usr.user_name)
    log.warning(f"ユーザーID: {usr.user_id} は認証されませんでした。")
    return None


class SaySmallDedaultHandler(tornado.web.RequestHandler):
    """リクエストハンドラーフック

    アラートフィルター用のリクエストハンドラーフック。

    """

    def set_nonce(self) -> str:
        """CSP スクリプトヘッダー生成

        CSP レスポンスヘッダーの script-src ディレクティブを生成する

        Return
        ------
        str
            nonce_id
        """
        nonce_id: str = urlsafe_b64encode(random_string(10).encode()).decode()
        self.set_header("Content-Security-Policy", f"script-src 'self' 'nonce-{nonce_id}'")

        return nonce_id

    def update_auth_cookie(self, info: AuthInfo):
        info.update()
        self.set_signed_cookie(
            COOKIE_NAME,
            info.model_dump_json(),
            expires_days=EXPIRE_DAY,
            samesite="strict",
            httponly=True,
            secure=SayTorConfig().use_ssl(),
        )

    def init_auth_cookie(self):
        self.clear_cookie(
            COOKIE_NAME,
            samesite="strict",
            httponly=True,
            secure=SayTorConfig().use_ssl(),
        )
        get_logger(__name__).detail(f"cookie {COOKIE_NAME} をクリア")


class DefaultBase(SaySmallDedaultHandler):
    def get_current_user(self) -> tuple[str]:
        log = get_logger(__name__)
        result: tuple[str] = None

        i_dat = self.get_signed_cookie(COOKIE_NAME)

        if i_dat is not None:
            try:
                data = json.loads(i_dat)
                info = AuthInfo(**data)
                if info.is_expire():
                    # 期限切れ
                    log.debug("cookie の情報は破棄されました。")
                    self.init_auth_cookie()
                else:
                    result = (info.user_id, info.user_name)
                    self.update_auth_cookie(info)
                    log.debug("cookie の内部時刻を更新しました。")

            except Exception as ex:
                log.warning(f"識別情報エラーです。理由: {str(ex)}, データ: {i_dat}")
        elif hasattr(self, "auth_user_info"):
            log.detail("cookie 外情報取得")
            result = self.auth_user_info
        else:
            log.debug(f"識別情報はありません。")

        return result

    def set_current_user(self, user_id: str, user_name: str):
        # cookie が取得できないときの補助
        self.auth_user_info = (user_id, user_name)

        # 認証用データを保存
        info = AuthInfo(
            user_id=user_id,
            user_name=user_name,
            expire_time=60 * 60,
        )
        self.update_auth_cookie(info)

    def error_render(self, status_code: int, error_title: str, error_message: str):
        """エラーページ遷移

        エラーページに遷移させる

        Parameters
        ----------
        status_code:int
            HTTP ステータスコード
        error_title:str
            画面のタイトル
        error_message:str
            エラーメッセージ本体
        """
        params: dict[str, object] = DefaultBase.get_common_args()
        params["statusCode"] = status_code
        params["errorTitle"] = error_title
        params["errorMessage"] = error_message

        self.render("error.html", **params)

    def write_error(self, status_code, **kwargs):
        """レスポンスエラー処理

        レスポンス返却時のエラー処理。
        既存メソッドのオーバーライド

        Parameters
        ----------
        status_code:int
            HTTPステータスコード
        kwargs:dict
            返却値のパラメータ
        """

        if "exc_info" in kwargs:
            msg = str(kwargs["exc_info"][1])
        else:
            msg = ""

        params = {
            "sysVersion": "",
            "statusCode": status_code,
            "errorTitle": self._reason,
            "errorMessage": msg,
        }
        params["isProduct"] = 0 if is_trial() != 0 else 1
        params["leaseDay"] = get_remain_day()

        self.render("error.html", **params)

    def err_response(self, status_code: int, msg_type: str, msg_body: str, detail_type: int, detail=None):
        """エラー返却メソッド

        共通形式のレスポンスを返却するメソッド

        Parameters
        ----------
        status_code:int
            HTTP ステータスコード
        msg_type:str
            メッセージタイプ
        msg_body:str
            メッセージ本文
        detail_type:int
            詳細データ形式
        detail
            詳細データ(省略可)
        """
        res = SysErrorResponse(
            message_type=msg_type,
            message=msg_body,
            detail_type=detail_type,
            detail=detail,
        )

        self.set_status(status_code)
        self.write(res.model_dump())
        self.finish()

    @classmethod
    def get_common_args(cls) -> dict[str, object]:
        cfg = SayTorConfig()
        result: dict[str, object] = {}
        result["sysVersion"] = cfg.get_config("sysVersion")
        result["isProduct"] = 0 if is_trial() != 0 else 1
        result["leaseDay"] = get_remain_day()

        return result


class BdBase(DefaultBase):
    def _request_auth(self):
        """ユーザー認証要求
        UNAUTHORIZED を返却する
        """
        if self.request.method == "GET":
            ua: str = self.request.headers.get("user-agent")
            if ua:
                for id in BROWSER_IDENTITY:
                    if id in ua.lower():
                        # クライアントブラウザなのでログインページに飛ばす
                        self.redirect("/login")
                        return

        # API なので json を返却
        self.set_header("WWW-Authenticate", "Basic realm=JSL")
        self.set_status(HTTPStatus.UNAUTHORIZED)
        raise tornado.web.Finish()

    async def prepare(self):
        log = get_logger(__name__)

        # アクセスログ
        log.info(f"{self.request.remote_ip} から {self.request.uri} へのリクエスト")

        # 認証済み cookie の確認
        if self.get_current_user():
            log.debug("認証済み")
            return

        auth_header = self.request.headers.get("Authorization")
        if auth_header is None:
            log.warning("認証リクエストがありません。")
            self._request_auth()
            return

        if not auth_header.startswith("Basic "):
            log.warning("不明な認証リクエストです。")
            self._request_auth()
            return

        auth_decoded = get_bytes(auth_header[6:]).decode()
        username, password = auth_decoded.split(":", 2)

        info: tuple[str] = await api_auth(username, password)
        if info is None:
            self._request_auth()
            return

        # API 認証でなければ認証情報を一時保存する
        if info[0] != SYSTEM_USER_ID:
            self.set_current_user(*info)

    def get_payload(self) -> dict[str, str]:
        log = get_logger(__name__)
        data: dict[str, str] = {}
        try:
            data = json_decode(self.request.body)
        except json.decoder.JSONDecodeError as ex:
            log.warning(f"リクエスト本文のデコードに失敗しました。({str(ex)})")
            data = None
        except Exception:
            log.warning("デコード処理でエラーが発生しました。", exc_info=True)
            data = None

        return data

    def write(self, data):
        """レスポンス処理

        dict データを JSON にして返却する。
        既存メソッドのオーバーライド

        Parameters
        ----------
        data:dict
            返却する dict。list を設定してはいけない。
        """
        if self._finished:
            # 書き込み前に finish している
            raise RuntimeError("Cannot write() after finish()")

        if not isinstance(data, (bytes, unicode_type, dict)):
            # 所定の形式ではない
            message = "write() only accepts bytes, unicode, and dict objects"
            if isinstance(data, list):
                message += ". Lists not accepted for security reasons; see http://www.tornadoweb.org/en/stable/web.html#tornado.web.RequestHandler.write"
            raise TypeError(message)

        if isinstance(data, dict):
            # JSON データの場合はヘッダーも付与
            response = json_encode(data)
            self.set_header("Content-Type", "application/json")
        else:
            response = data

        # 返却値をエスケープ処理
        response = utf8(response)
        self._write_buffer.append(response)
