import re
from http import HTTPStatus

from app import get_logger
from app.db import get_ses_maker
from app.crud.configure import delete, save
from app.enum import ResponseDataType
from app.entity.configure import Configure
from app.router import BdBase, NO_PATH_PARAMS, GET_UID, ITEM_NOT_FOUND_ERROR
from app.service.configure import SayTorConfig
from app.util.env_config import set_env, delete_env

# 環境変数のみ許可される設定キー名
REJECT_CONFIG_KEYS = re.compile(
    "rootUrl|debugMode|port|database|conn_args|sysVersion|_dir|_keys*|overwrite_environ|_limit", re.I
)


class ServerConfigHandler(BdBase):
    async def get(self, *args, **kwargs):
        """設定取得

        設定取得API

        Request
        ----------
        keys:list[str]
            設定を取得するキー
        """
        NO_PATH_PARAMS(*args, **kwargs)
        tmp_keys = self.get_arguments("keys", True)
        keys = [i for i in tmp_keys if len(i) > 0]
        if len(keys) <= 0:
            self.err_response(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                msg_type="E001",
                msg_body="リクエストパラメータが正しくありません。",
                detail_type=ResponseDataType.d_null,
            )
            return

        result: dict = {}
        cfg = SayTorConfig()
        for k in keys:
            result[k] = cfg.get_config(k, None)

        self.set_status(HTTPStatus.OK)
        self.write(result)
        self.finish()

    async def post(self, *args, **kwargs):
        """設定保存

        設定保存API

        Parameters
        ----------
        dict[str,str]
            設定値。JSON
        """
        NO_PATH_PARAMS(*args, **kwargs)
        log = get_logger(__name__)
        data: dict = self.get_payload()
        if data is None:
            self.err_response(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                msg_type="E001",
                msg_body="リクエストパラメータが正しくありません。",
                detail_type=ResponseDataType.d_null,
            )
            return

        count: int = 0
        result: dict = {}
        cfg = SayTorConfig()
        uid = GET_UID(self.current_user)

        for k in data.keys():
            # API で設定できないキーでないか確認
            if REJECT_CONFIG_KEYS.search(k):
                log.info(f"{k} は、除外対象キーです。")
                continue

            v: str = str(data[k]) if data[k] is not None else None
            cfg.set_config(k, v)
            sv = cfg.get_raw_config(k)  # 保存用の値

            try:
                ret = None
                cfg_type: int = cfg.has_config(k)
                if cfg_type == 1:
                    set_env(k, sv)
                    ret = True
                else:
                    ent = Configure(ckey=k, cval=sv, update_by=uid)
                    async with get_ses_maker() as db:
                        ret = await save(db, ent)

                if ret:
                    count += 1
            except Exception:
                log.exception(f"設定 {k} の保存に失敗しました。")

        result["code"] = HTTPStatus.CREATED
        result["count"] = count

        self.set_status(HTTPStatus.CREATED)
        self.write(result)
        self.finish()

    async def delete(self, *args):
        """設定削除

        設定削除API

        Parameters
        ----------
        list[str]
            設定名
        """
        log = get_logger(__name__)
        key = ""
        if len(args) >= 1:
            key = args[0]

        if len(key) <= 0:
            self.err_response(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                msg_type="E001",
                msg_body="リクエストパラメータが正しくありません。",
                detail_type=ResponseDataType.d_null,
            )
            return

        # API で設定できないキーでないか確認
        if REJECT_CONFIG_KEYS.search(key):
            self.err_response(
                status_code=HTTPStatus.FORBIDDEN,
                msg_type="E021",
                msg_body=f"{key} は設定することができません。",
                detail_type=ResponseDataType.d_null,
            )
            return

        count: int = 0
        result: dict = {}
        cfg = SayTorConfig()
        cfg_type: int = cfg.has_config(key)
        if cfg_type not in [1, 2]:
            log.warning(f"{key} は、永続データのキーではありません。")
            raise ITEM_NOT_FOUND_ERROR

        cfg.delete_config(key)
        try:
            ret = None
            if cfg_type == 1:
                delete_env(key)
                ret = True
            elif cfg_type == 2:
                async with get_ses_maker() as db:
                    ret = await delete(db, key)

            if ret:
                count += 1
        except Exception:
            log.exception(f"{key} の設定を削除できませんでした。")

        result["code"] = HTTPStatus.CREATED
        result["count"] = count

        self.set_status(HTTPStatus.CREATED)
        self.write(result)
        self.finish()
