from http import HTTPStatus

from pydantic import ValidationError

# from memory_profiler import profile

from app import get_logger, to_int
from app.router import BdBase, ITEM_NOT_FOUND_ERROR, NO_PATH_PARAMS, SERVER_RUNTIME_ERROR, ABEND_SAVE_ERROR, GET_UID
from app.db import get_ses_maker
from app.crud.mail_filter import get, create, delete, update, get_count, get_list
from app.enum import ResponseDataType
from app.entity import get_validate_errors
from app.entity.mail_filter import MailFilter, create_empty_filter
from app.service.filter_cache import MailFilterCache
from app.service.license import get_max_filter_count


class FilterListHandler(BdBase):
    """フィルター一覧画面ハンドラー"""

    async def get(self):
        log = get_logger(__name__)

        # 取得ページ
        pages: list[str] = self.get_arguments("page", True)
        page: int = 1
        if len(pages) > 0:
            log.detail(f"ページ指定:{str(pages)}")
            page = to_int(pages[0], 1)

        # 設定取得
        from app.service.configure import SayTorConfig

        cfg = SayTorConfig()
        list_row_cnt: int = to_int(cfg.get_config("page_row", "10"), 10)
        off_set: int = (page - 1) * list_row_cnt
        log.debug(f"ページ={page}, オフセット={off_set}, リストの最大件数={list_row_cnt}")

        async with get_ses_maker() as db:
            # 指定件数のデータ取得
            list_item: list[MailFilter] = await get_list(db, lmt=list_row_cnt, off_s=off_set)
            log.detail(f"\t取得データ件数={len(list_item)}")

            # データ件数の取得
            total_cnt: int = await get_count(db)

        # 画面パラメータ設定
        params: dict = BdBase.get_common_args()

        curr: tuple[str] = self.get_current_user()
        if curr is not None:
            params["current_id"] = curr[0]
            params["current_nm"] = curr[1]
        else:
            params["current_id"] = ""
            params["current_nm"] = ""

        params["listItems"] = list_item
        params["rowCnt"] = total_cnt
        params["maxCnt"] = get_max_filter_count()

        # ページ処理用
        params["currentPage"] = page
        params["lineCnt"] = list_row_cnt

        # レンダリング
        self.render("filter/list.html", **params)


class FilterItemHandler(BdBase):
    async def get(self, filter_id=None):
        """GETメソッド

        メールフィルターの取得メソッド
        取得後、設定画面に遷移する

        Parameters
        ----------
        args:list[str]
            パスパラメータ
        """
        # 画面パラメータ設定
        params: dict = BdBase.get_common_args()
        params["nonce_id"] = self.set_nonce()

        curr: tuple[str] = self.get_current_user()
        if curr is not None:
            params["current_id"] = curr[0]
            params["current_nm"] = curr[1]
        else:
            params["current_id"] = ""
            params["current_nm"] = ""

        # 複製かどうか
        copy_flg = self.get_argument("method", "", True)

        if filter_id is not None:
            # フィルター ID がある場合はデータ取得
            async with get_ses_maker() as db:
                item: list[MailFilter] = await get(db, int(filter_id))

            if len(item) <= 0:
                self.error_render(
                    status_code=HTTPStatus.NOT_FOUND,
                    error_title="システムエラー：データはありません。",
                    error_message=f"ID:{filter_id} のメールフィルターが見つかりませんでした。",
                )
                return
            params["filterItem"] = item[0]
            if copy_flg == "cp":
                params["filterItem"].filter_id = 0  # ID を消す
        else:
            params["filterItem"] = create_empty_filter()

        # レンダリング
        self.render("filter/filter.html", **params)

    # @profile(precision=4)
    async def put(self, filter_id=None):
        if filter_id is None:
            raise ITEM_NOT_FOUND_ERROR

        log = get_logger(__name__)

        data: dict = self.get_payload()
        if data is None:
            self.err_response(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                msg_type="E001",
                msg_body="リクエストパラメータが正しくありません。",
                detail_type=ResponseDataType.d_null,
            )
            return

        try:
            f = MailFilter(**data)
            f.penetrate_filter_id()
            f.penetrate_update_by(GET_UID(self.get_current_user()))
        except ValidationError as ex:
            self.err_response(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                msg_type="E001",
                msg_body="リクエストパラメータが正しくありません。",
                detail_type=ResponseDataType.d_hash,
                detail=get_validate_errors(ex),
            )
            return
        except Exception:
            log.exception("リクエストパラメータの形式が正しくありません。")
            raise ABEND_SAVE_ERROR

        # 保存時のアクション補正(受信時に勝手に変換されてしまうため)
        for a in f.filter_actions:
            a.act_data = a.act_data.replace("+", "%20")

        # UPDATE(実際は DELETE / INSERT)
        async with get_ses_maker() as db:
            try:
                result = await update(db, f)
            except Exception:
                log.exception(f"フィルター ID: {f.filter_id} のメールフィルター更新でエラーが発生しました。")
                raise SERVER_RUNTIME_ERROR

        if result is None:
            log.error(f"フィルター名: {f.filter_name} のメールフィルター更新に失敗しました。")
            raise ABEND_SAVE_ERROR
        else:
            await MailFilterCache().reload()
            log.info(f"フィルター ID: {result['filter_id']} のメールフィルター更新が完了しました。")
            self.set_status(HTTPStatus.CREATED)
            self.write(result)
            self.finish()

    # @profile(precision=4)
    async def post(self, *args, **kwargs):
        NO_PATH_PARAMS(*args, **kwargs)
        log = get_logger(__name__)

        data: dict = self.get_payload()
        if data is None:
            log.warning("リクエストの本文がありません。")
            self.err_response(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                msg_type="E001",
                msg_body="リクエストパラメータが正しくありません。",
                detail_type=ResponseDataType.d_null,
            )
            return

        try:
            f = MailFilter(**data)
            f.penetrate_update_by(GET_UID(self.get_current_user()))
        except ValidationError as ex:
            self.err_response(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                msg_type="E001",
                msg_body="リクエストパラメータが正しくありません。",
                detail_type=ResponseDataType.d_hash,
                detail=get_validate_errors(ex),
            )
            return
        except Exception:
            log.exception("リクエストパラメータの形式が正しくありません。")
            raise SERVER_RUNTIME_ERROR

        # 保存時のアクション補正(受信時に勝手に変換されてしまうため)
        for a in f.filter_actions:
            a.act_data = a.act_data.replace("+", "%20")

        # 登録可能件数の確認
        async with get_ses_maker() as db:
            try:
                maxCnt = await get_count(db)
            except Exception:
                log.exception("メールフィルターの登録確認時にエラーが発生しました。")
                raise SERVER_RUNTIME_ERROR

        curCnt = get_max_filter_count()
        if maxCnt >= curCnt:
            msg = f"メールフィルターの登録可能な件数の上限を超えました。最大 {maxCnt} 件までです。"
            log.warning(msg)
            self.err_response(
                status_code=HTTPStatus.BAD_REQUEST,
                msg_type="E300",
                msg_body=msg,
                detail_type=ResponseDataType.d_null,
            )
            return

        async with get_ses_maker() as db:
            try:
                result = await create(db, f)
            except Exception:
                log.exception(f"フィルター名:{f.filter_name} 優先度: {f.priority} のメールフィルター登録でエラーが発生しました。")
                raise SERVER_RUNTIME_ERROR

        if result is None:
            log.error(f"フィルター名: {f.filter_name} のメールフィルター登録に失敗しました。")
            raise ABEND_SAVE_ERROR
        else:
            await MailFilterCache().reload()
            log.info(f"フィルター ID: {result['filter_id']} でメールフィルターの新規登録が完了しました。")
            self.set_status(HTTPStatus.CREATED)
            self.write(result)
            self.finish()

    # @profile(precision=4)
    async def delete(self, filter_id=None):
        if filter_id is None:
            raise ITEM_NOT_FOUND_ERROR

        log = get_logger(__name__)

        # データベースの削除
        async with get_ses_maker() as db:
            try:
                # 派生データも一括で消してくれる
                result = await delete(db, int(filter_id))
            except Exception:
                log.exception(f"フィルター ID: {filter_id} のメールフィルター削除でエラーが発生しました。")
                raise SERVER_RUNTIME_ERROR

        if result is None:
            log.error(f"フィルター ID: {filter_id} のメールフィルター削除に失敗しました。")
            raise ABEND_SAVE_ERROR
        else:
            await MailFilterCache().reload()
            log.info(f"フィルター ID: {filter_id} のメールフィルターを削除しました。")
            self.set_status(HTTPStatus.NO_CONTENT)
            # self.write(result)    # NO_CONTENT したら BODY は None じゃないとダメ
            self.finish()
