from http import HTTPStatus

from pydantic import ValidationError

from app import get_logger
from app.db import get_ses_maker
from app.enum import MailProtocol, ResponseDataType
from app.crud.mail_server import get, create, update, delete, get_as_protocol
from app.router import BdBase, ITEM_NOT_FOUND_ERROR, SERVER_RUNTIME_ERROR, ABEND_SAVE_ERROR, NO_PATH_PARAMS, GET_UID
from app.entity import get_validate_errors
from app.entity.mail_server import MailServer
from app.service.pop import has_server, start_server, stop_server
from app.service.mail_send import stop_retry_send, start_retry_send


class MailServerHandler(BdBase):
    async def get(self, host=None, port=None):
        if host is None or port is None:
            raise ITEM_NOT_FOUND_ERROR

        log = get_logger(__name__)

        async with get_ses_maker() as db:
            try:
                res = await get(db, host, port)
            except Exception:
                log.exception(f"メールサーバー {host}:{port} の設定取得でエラーが発生しました。")
                raise SERVER_RUNTIME_ERROR

        if res is None:
            log.warning(f"メールサーバー {host}:{port} の設定取得に失敗しました。")
            raise ITEM_NOT_FOUND_ERROR
        else:
            self.write(res.model_dump())
            self.flush()

    async def delete(self, host=None, port=None):
        log = get_logger(__name__)

        if host is None or port is None:
            log.warning(f"必須パラメータがありません。リクエストホスト名:{host} ポート番号:{port}")
            raise ITEM_NOT_FOUND_ERROR

        # ポーリングの停止(あれば)
        if await has_server(host, port):
            if not await stop_server(host, port):
                log.error(f"メールサーバー {host}:{port} の定期受信が停止できませんでした。")
                self.err_response(
                    status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                    msg_type="E201",
                    msg_body=f"メールサーバー {host}:{port} の定期受信が停止できません。",
                    detail_type=ResponseDataType.d_null,
                )
                return

        # データベースの削除
        async with get_ses_maker() as db:
            try:
                result = await delete(db, host=host, port=port)
            except Exception:
                log.exception(f"メールサーバー {host}:{port} の情報削除でエラーが発生しました。")
                raise SERVER_RUNTIME_ERROR

        if result is None:
            log.error(f"メールサーバー {host}:{port} の情報削除に失敗しました。")
            raise ABEND_SAVE_ERROR
        elif result["protocol"] == MailProtocol.smtp:
            # ポーリングの停止
            await stop_retry_send()

        log.info(f"メールサーバー {host}:{port} の情報は削除されました。")
        self.set_status(HTTPStatus.NO_CONTENT)  # NO_CONTENT したら BODY は None じゃないとダメ
        self.finish()

    async def post(self, *args, **kwargs):
        NO_PATH_PARAMS(*args, **kwargs)
        log = get_logger(__name__)

        data: dict = self.get_payload()
        if data is None:
            log.warning("リクエスト本文がありません。")
            self.err_response(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                msg_type="E001",
                msg_body="リクエストパラメータが正しくありません。",
                detail_type=ResponseDataType.d_null,
            )
            return

        try:
            u = MailServer(**data)
            u.update_by = GET_UID(self.get_current_user())
        except ValidationError as ex:
            self.err_response(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                msg_type="E001",
                msg_body="リクエストパラメータが正しくありません。",
                detail_type=ResponseDataType.d_hash,
                detail=get_validate_errors(ex),
            )
            return
        except Exception:
            log.exception("リクエストパラメータの形式が正しくありません。")
            raise SERVER_RUNTIME_ERROR

        # 既に起動している場合はエラー
        if await has_server(u.host, u.port):
            log.warning(f"既にメールサーバー {u.host}:{u.port} は定期受信が実行中です。")
            self.err_response(
                status_code=HTTPStatus.BAD_REQUEST,
                msg_type="E202",
                msg_body=f"メールサーバー {u.host}:{u.port} は既に定期実行されています。",
                detail_type=ResponseDataType.d_null,
            )
            return

        # 既に送信設定がある場合、送信設定の追加は NG
        if u.protocol == MailProtocol.smtp:
            async with get_ses_maker() as db:
                svr = await get_as_protocol(db, MailProtocol.smtp)

            if len(svr) > 0:
                log.error(f"既に送信メールサーバー {svr[0].host}:{svr[0].port} が登録されています。リクエストは破棄されます。")
                self.err_response(
                    status_code=HTTPStatus.BAD_REQUEST,
                    msg_type="E203",
                    msg_body="SMTP サーバーは既に登録されています。",
                    detail_type=ResponseDataType.d_null,
                )
                return

        async with get_ses_maker() as db:
            try:
                result = await create(db, u)
            except Exception:
                log.exception(f"メールサーバー {u.host}:{u.port} の情報登録でエラーが発生しました。")
                raise SERVER_RUNTIME_ERROR

        if result is None:
            log.error(f"メールサーバー {u.host}:{u.port} の情報登録に失敗しました。")
            raise ABEND_SAVE_ERROR

        if u.protocol == MailProtocol.pop3:
            # POP3 のポーリング開始
            if not await start_server(u):
                log.error(f"メールサーバー {u.host}:{u.port} の定期受信が開始できませんでした。")
                self.err_response(
                    status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                    msg_type="E204",
                    msg_body=f"メールサーバー {u.host}:{u.port} の定期実行に失敗しました。",
                    detail_type=ResponseDataType.d_null,
                )
                return
            log.info(f"メールサーバー {u.host}:{u.port} の定期受信が開始されました。")
        else:
            # ポーリングの開始
            await start_retry_send(u)
            log.info(f"送信用メールサーバー {u.host}:{u.port} の情報登録が完了しました。")

        self.set_status(HTTPStatus.CREATED)
        self.write(result)
        self.finish()

    async def put(self, *args, **kwargs):
        NO_PATH_PARAMS(*args, **kwargs)
        log = get_logger(__name__)

        data: dict = self.get_payload()
        if data is None:
            log.warning("リクエスト本文がありません。")
            self.err_response(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                msg_type="E001",
                msg_body="リクエストパラメータが正しくありません。",
                detail_type=ResponseDataType.d_null,
            )
            return

        try:
            u = MailServer(**data)
            u.update_by = GET_UID(self.get_current_user())
        except ValidationError as ex:
            self.err_response(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                msg_type="E001",
                msg_body="リクエストパラメータが正しくありません。",
                detail_type=ResponseDataType.d_hash,
                detail=get_validate_errors(ex),
            )
            return
        except Exception:
            log.exception("リクエストパラメータの形式が正しくありません。")
            raise SERVER_RUNTIME_ERROR

        # POP3 で起動していない場合はエラー
        if u.protocol == MailProtocol.pop3:
            if not await has_server(u.host, u.port):
                log.warning(f"メールサーバー {u.host}:{u.port} の定期受信が実行されていません。")
                self.err_response(
                    status_code=HTTPStatus.BAD_REQUEST,
                    msg_type="E205",
                    msg_body=f"メールサーバー {u.host}:{u.port} が定期実行されていません。",
                    detail_type=ResponseDataType.d_null,
                )
                return

        async with get_ses_maker() as db:
            try:
                result = await update(db, u)
            except Exception:
                log.exception(f"メールサーバー {u.host}:{u.port} の情報更新でエラーが発生しました。")
                raise SERVER_RUNTIME_ERROR

        if result is None:
            log.error(f"メールサーバー {u.host}:{u.port} の情報更新に失敗しました。")
            raise ABEND_SAVE_ERROR

        # サービス再起動
        if u.protocol == MailProtocol.pop3:
            if not await stop_server(u.host, u.port):
                log.warning(f"メールサーバー {u.host}:{u.port} の定期受信が停止できませんでした。")
                self.err_response(
                    status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                    msg_type="E201",
                    msg_body=f"メールサーバー {u.host}:{u.port} の定期実行が停止できません。",
                    detail_type=ResponseDataType.d_null,
                )
                return
            elif not await start_server(u):
                log.warning(f"メールサーバー {u.host}:{u.port} の定期受信が再起動できませんでした。")
                self.err_response(
                    status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                    msg_type="E204",
                    msg_body=f"メールサーバー {u.host}:{u.port} の定期実行に失敗しました。",
                    detail_type=ResponseDataType.d_null,
                )
                return
            log.info(f"メールサーバー {u.host}:{u.port} の定期受信は再起動しました。")
        else:
            # ポーリングの再起動(念のため)
            await stop_retry_send()
            await start_retry_send(u)
            log.info(f"送信メールサーバー {u.host}:{u.port} の情報更新が完了しました。")

        self.set_status(HTTPStatus.OK)
        self.write(result)
        self.finish()
