from http import HTTPStatus

from app import get_logger, COOKIE_NAME
from app.db import get_ses_maker
from app.crud.users import change_password, get
from app.router import BdBase, SERVER_RUNTIME_ERROR, GET_UID
from app.entity.users import ChangePassword
from app.enum import ResponseDataType


class UsersHandler(BdBase):
    async def get(self, uid):
        log = get_logger(__name__)

        async with get_ses_maker() as db:
            try:
                res = await get(db, uid)
            except Exception:
                log.exception("ユーザー情報の取得でエラーが発生しました。")
                raise SERVER_RUNTIME_ERROR

        if res is None:
            log.warning(f"ユーザー ID: {uid} の情報はありません。")
            self.err_response(
                status_code=HTTPStatus.NOT_FOUND,
                msg_type="W001",
                msg_body="データはありませんでした。",
                detail_type=ResponseDataType.d_null,
            )
            return
        else:
            log.info(f"ユーザー ID: {uid} の情報を取得しました。")
            self.set_status(HTTPStatus.OK)
            self.write(res.model_dump())
            self.flush()


class ChangePassHandler(BdBase):
    async def get(self):
        """GETメソッド

        パスワード変更
        パスワード変更画面に遷移する
        """
        # 画面パラメータ設定
        params: dict = BdBase.get_common_args()

        curr: tuple[str] = self.get_current_user()
        params["current_id"] = curr[0]
        params["current_nm"] = curr[1]

        # レンダリング
        self.render("change_pass.html", **params)

    async def put(self, uid):
        log = get_logger(__name__)
        data: dict = self.get_payload()
        if data is None:
            log.warning("リクエスト本文がありません。")
            self.err_response(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                msg_type="E001",
                msg_body="リクエストパラメータが正しくありません。",
                detail_type=ResponseDataType.d_null,
            )
            return

        try:
            u = ChangePassword(user_id=uid, update_by=GET_UID(self.get_current_user()), **data)
        except ValueError as ex:
            self.err_response(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                msg_type="E001",
                msg_body="リクエストパラメータが正しくありません。",
                detail_type=ResponseDataType.d_primitive,
                detail=str(ex),
            )
            return
        except Exception:
            log.exception("リクエストパラメータの形式が正しくありません。")
            raise SERVER_RUNTIME_ERROR

        async with get_ses_maker() as db:
            try:
                result = await change_password(db, u)
            except Exception:
                log.exception("パスワードの保存でエラーが発生しました。")
                raise SERVER_RUNTIME_ERROR

        if result is None:
            log.warning("パスワードの保存に失敗しました。")
            self.err_response(
                status_code=HTTPStatus.BAD_REQUEST,
                msg_type="E100",
                msg_body="パスワードの更新に失敗しました。",
                detail_type=ResponseDataType.d_primitive,
                detail="パスワードは変更されませんでした。",
            )
            return
        else:
            self.init_auth_cookie()
            log.info(f"ユーザー ID: {uid} のパスワードが変更されました。")
            self.set_status(HTTPStatus.OK)
            self.write(result)
            self.finish()
