import os
import re
import json
import subprocess
from urllib.parse import unquote

import psutil
from tornado.httpclient import HTTPClientError, HTTPRequest, HTTPResponse, AsyncHTTPClient
from chardet import detect

# from memory_profiler import profile

from app import get_logger
from app.enum import ActionType
from app.entity.action_data import SendMailActionData, WebhookActionData, CommandActionData
from app.entity.mail_filter import FilterAction
from app.service.send_eml import SendMailMsg
from app.service.mail_send import new_sender


class BaseAction(object):
    """アクションベース

    アクションのベースクラス

    """

    def __init__(self, act: str):
        """初期化処理

        当該アクションの初期化

        Parameters
        ----------
        act:str
            アクションデータ JSON 文字列
        """
        # パラメータチェック
        if not act:
            raise ValueError("nothing action data")

        # アクションデータを格納
        self.raw = act
        self.param = json.loads(unquote(self.raw))

        # 変数データを格納
        self.vals: dict[str, str] = {}

    def set_values(self, vals: dict[str, str]):
        """動的パラメータ値取得

        パラメータ変数の値を取得する

        Parameters
        ----------
        vals:dict[str, str]
            変数名と値の dict
        """
        self.vals.update(vals)

    def replace_values(self, line: str) -> str:
        """動的パラメータ置換

        与えられた文字列を変数で置換する

        Parameters
        ----------
        vals:dict[str, str]
            変数名と値の dict
        """
        result: str = line
        for k, v in self.vals.items():
            result = result.replace(f"$({k})", v)

        return result

    async def execute(self) -> bool:
        """アクション実行

        Return
        ------
        bool
            処理成功時、True
        """
        raise NotImplementedError("'execute' method not implement.")


class SendMailAction(BaseAction):
    def __init__(self, act: str):
        # 初期化
        super().__init__(act)
        self.data = SendMailActionData(**self.param)

    # @profile(precision=4)
    async def execute(self) -> bool:
        log = get_logger(__name__)
        result: bool = False

        # 改行補正(これを入れないと送信メールが改行されない)
        b_data = self.data.body.replace("\n", "\r\n")

        log.info("フィルターアクション「メール送信」開始。")
        msg = SendMailMsg(
            subject=self.replace_values(self.data.subject),
            body=self.replace_values(b_data),
            addr_to=[self.replace_values(addr) for addr in self.data.to.split(",")],
        )

        svr = await new_sender()
        if svr:
            log.info("メール送信開始。")
            result = svr.send(msg)
        else:
            log.warning("\t送信設定がありませんでした。メール送信はキャンセルされます。")

        log.info("フィルターアクション「メール送信」終了。")
        return result


class WebhookAction(BaseAction):
    def __init__(self, act: str):
        # 初期化
        super().__init__(act)
        self.data = WebhookActionData(**self.param)

    # @profile(precision=4)
    async def execute(self) -> bool:
        log = get_logger(__name__)

        # ヘッダーの事前分解
        headers: dict[str, str] = {}
        if self.data.header:
            for line in self.data.header.split("\n"):
                if line.count("=") <= 0:
                    log.detail(f"\t無効なリクエストヘッダー=[{line}]")
                    continue
                apply_val: str = self.replace_values(line)
                (key, val) = apply_val.split("=", 1)
                headers[key.strip()] = val.strip()

        cli = AsyncHTTPClient()
        req = HTTPRequest(
            url=self.replace_values(self.data.url),
            method=self.data.method,
            headers=headers,
            validate_cert=False,  # SSL 証明書を検証しない
        )

        # GET/DELETE には本文をつけない
        if self.data.method.upper() not in ["GET", "DELETE"]:
            req.body = self.replace_values(self.data.body)

        result = False

        try:
            log.info(f"フィルターアクション「Webhook」を {self.data.url} に {self.data.method} メソッドで開始。")
            res: HTTPResponse = await cli.fetch(req)
            log.info(f"フィルターアクション「Webhook」終了。HTTP レスポンス: {res.code}, レスポンスボディ: {res.body.decode()}")
            result = True
        except (ConnectionRefusedError, ConnectionResetError):
            log.error(f"フィルターアクション「Webhook」を {self.data.url} に実行しましたが、接続でエラーが発生しました。")
        except HTTPClientError as ex:
            log.error(
                f"フィルターアクション「Webhook」を {self.data.url} に {self.data.method} メソッドで行い、エラーが発生しました。コード: {ex.code}、メッセージ: {str(ex.response)}"
            )
        except Exception as ex:
            log.exception(
                f"フィルターアクション「Webhook」を {self.data.url} に {self.data.method} メソッドで行い、エラーが発生しました。エラー理由: {str(ex)}"
            )

        return result


class CommandAction(BaseAction):
    def __init__(self, act: str):
        # 初期化
        super().__init__(act)
        self.data = CommandActionData(**self.param)

    # @profile(precision=4)
    async def execute(self) -> bool:
        log = get_logger(__name__)
        result = False

        rej_ext = None
        ext: str = os.environ.get("PATHEXT")
        if ext:
            rej_ext = re.compile(ext.replace(";", "|"), re.I)

        # 拡張子対応
        get_proc_name = lambda x: x if rej_ext is None else rej_ext.sub("", x)

        command_line: str = self.replace_values(self.data.command)
        b_name = os.path.basename(command_line)
        command_line += " " + self.replace_values(self.data.options)

        cwd = self.replace_values(self.data.work_dir) if len(self.data.work_dir) > 0 else None

        # プロセス多重実行の確認
        if not self.data.multi:
            my_proc = [
                p.pid
                for p in psutil.process_iter(["name"])
                if get_proc_name(p.info["name"]).casefold() == get_proc_name(b_name).casefold()
            ]
            if len(my_proc) > 0 and not self.data.multi:
                log.warning(f"コマンド {b_name} は既に実行中です。")
                return False

        log.debug(f"コマンドライン: '{command_line}'")

        try:
            log.info("フィルターアクション「コマンド実行」開始。")

            # コマンドの実行
            ret = subprocess.run(command_line, capture_output=True, cwd=cwd)
            stdout_msg = ""
            stderr_msg = ""
            if ret is not None:
                data: bytes = ret.stdout
                if len(data) > 0:
                    log.detail(f"\tSTDOUT[{data}]")
                    enc = detect(data)
                    if enc["encoding"] is None:
                        enc = {"encoding": "UTF-8"}
                    stdout_msg = data.decode(encoding=enc["encoding"], errors="ignore")
                    enc.clear()
                    enc = None

                data: bytes = ret.stderr
                if len(data) > 0:
                    log.detail(f"\tSTDERR[{data}]")
                    enc = detect(data)
                    if enc["encoding"] is None:
                        enc = {"encoding": "UTF-8"}
                    stderr_msg = data.decode(encoding=enc["encoding"], errors="ignore")
                    enc.clear()
                    enc = None

            # 実行結果
            log.info(
                f"フィルターアクション「コマンド実行」完了。\n\tコマンド: {b_name}\n\tEXIT ステータス={ret.returncode}\n\t 出力結果='{stdout_msg}'\n\t標準エラー='{stderr_msg}'"
            )
            result = True
        except Exception:
            log.exception(f"フィルターアクション「コマンド実行」でエラーが発生しました。コマンド: {b_name}")

        return result


def get_new_action(act: FilterAction) -> BaseAction:
    """アクションインスタンス生成

    タイプに応じてアクション実行用インスタンスを作成する

    Parameters
    ----------
    act:FilterAction
        フィルターに設定されているアクションエンティティ

    Return
    ------
    BaseAction
        タイプに応じたアクションインスタンス、タイプ不明の場合は None
    """
    result: BaseAction = None

    if act.act_type == ActionType.action_mail:
        result = SendMailAction(act.act_data)
    elif act.act_type == ActionType.action_webhook:
        result = WebhookAction(act.act_data)
    elif act.act_type == ActionType.action_command:
        result = CommandAction(act.act_data)

    return result
