from app.enum import FileType
from app.util.archive import BdCabinet, BdZip
from app.util.file_util import check_file_type, is_log_file, get_temp_dir, peaceably_delete
from app.service.configure import SayTorConfig


class AttachFiles:
    """添付ファイル操作クラス

    添付ファイル用に複数ファイルを管理する。

    """

    def __init__(self, lst: list[str] = []):
        """初期化処理

        クラスの初期化

        Parameters
        ----------
        lst:list[str]
            管理対象のファイルパス
        work_dir:str
            伸張用一時ディレクトリ
        """
        self.files: list[str] = []
        self.arch: list[tuple[any]] = []
        self.mod: list[str] = []
        self.conf = SayTorConfig()

        # ファイルを管理下に追加
        for f in lst:
            self.append(f)

    def __del__(self):
        """デストラクタ"""
        self.files.clear()
        self.arch.clear()
        self.mod.clear()
        self.files = None
        self.arch = None
        self.mod = None
        self.conf = None

    def file_count(self) -> int:
        """管理ファイル数返却

        管理しているファイル数を返す

        Return
        ----------
        int
            管理ファイル数
        """
        return len(self.files) + len(self.arch) + len(self.mod)

    def append(self, file_path: str):
        """管理ファイル追加

        管理するファイルのパスを追加する。
        圧縮ファイルの場合はチェック内容と合わせた結果を追加する。

        Parameters
        ----------
        file_path:str
            対象ファイルのパス
        """
        t = check_file_type(file_path)
        if t & FileType.archive:
            self.arch.append(tuple([file_path, t]))
        else:
            self.files.append(file_path)

    def modified(self, file_path: str):
        """修正済ファイル追加

        修正済み管理ファイルのパスを追加する。

        Parameters
        ----------
        file_path:str
            対象ファイルのパス
        """
        self.mod.append(file_path)

    def all_files(self) -> list[str]:
        """全管理ファイルパスリスト取得

        管理ファイルの全リストを返す

        Return
        ------
        list[str]
            全ファイルのパスリスト
        """
        result: list[str] = self.files + self.mod
        for f in self.arch:
            result.append(f[0])

        return result

    def pop_log(self) -> str:
        """ログファイル取得

        管理ファイル中からログファイルを取得する。
        ジェネレーター

        Return
        ------
        str
            ログファイルのパス
        """
        rmv_files: list[str] = []

        for f in self.files:
            if is_log_file(f):
                rmv_files.append(f)
                yield f

        for r in rmv_files:
            self.files.remove(r)

    def pop_arch(self):
        """圧縮ファイル取得

        管理ファイル中から圧縮ファイルを取得する。
        ジェネレーター

        Return
        ------
            圧縮ファイル管理用インスタンス。対象外の場合は None
        """
        rmv_files: list[str] = []

        for f in self.arch:
            if f[1] == FileType.cab:
                arch = BdCabinet(file_path=f[0])
            elif f[1] == FileType.zip:
                arch = BdZip(file_path=f[0])
            else:
                arch = None

            if arch is not None:
                rmv_files.append(f)

            yield arch

        for r in rmv_files:
            self.arch.remove(r)

    def to_zip(self, zip_name: str) -> bool:
        """ZIP圧縮

        管理ファイルを圧縮する

        Parameters
        ----------
        zip_name:str
            ZIPファイルのパス

        Return
        ------
        bool
            正常に作成された場合、True
        """
        if len(zip_name) <= 0:
            return False
        elif self.file_count() <= 0:
            peaceably_delete(zip_name)
            return False

        # ファイル名の確認
        file_name = zip_name if zip_name.lower().endswith(".zip") else zip_name + ".zip"

        # ZIP ファイルに圧縮する
        zip = BdZip(file_name)
        zip.files = self.all_files()
        return zip.compress()
