import os

from cryptography.fernet import Fernet

from app import get_logger, KEY_NAME
from app.util.env_config import set_env

"""
設定値の参照優先度は
    環境変数＞その他（設定ファイル＞データベース）
とする。
API経由では環境変数の値を永続保存しないので、再起動すると戻る
"""
## Global Variables ###########################################################


class SayTorConfig(object):
    def __new__(cls):
        if not hasattr(cls, "_instance"):
            cls._instance = super(SayTorConfig, cls).__new__(cls)
            cls._cfg: dict = {}
            cls.__init_key: bool = False
        return cls._instance

    def init_key(self):
        """暗号キー初期化

        暗号キーを初期化する。
        これを呼び出さないと暗号・復号は行えない。

        Return
        ------
        self
            自分自身
        """
        log = get_logger(__name__)
        if self.__init_key:
            log.warning("秘密鍵は既に初期化されています。")
            return self
        self.__init_key = True
        log.debug("秘密キー初期化実行。")

        key = self.get_config(KEY_NAME)
        if len(key) <= 0:
            key = Fernet.generate_key().decode()
            set_env(KEY_NAME, key)
            os.environ[KEY_NAME] = key
            log.info("秘密キーの生成完了。")

        return self

    def get_config(self, key: str, default=""):
        """設定取得

        設定を取得する

        Parameters
        ----------
        key: str
            設定名
        default: object
            設定が無い場合の初期値

        Returns
        -------
        object
            設定
        """
        ret = self.get_raw_config(key)

        # 初期値と復号化対応
        if self.is_crypt(key):
            if ret is not None:
                ret = self.decrypt(ret)

        if ret is None:
            ret = default
            self.set_config(key, default)

        return ret

    def get_raw_config(self, key: str):
        """設定生値取得

        未加工の設定値を取得する

        Parameters
        ----------
        key: str
            設定名

        Returns
        -------
        object
            設定値、または None
        """
        ret = None

        if key in os.environ:
            ret = os.environ[key]
        elif key in self._cfg:
            ret = self._cfg[key]

        return ret

    def set_config(self, key: str, val: str):
        """設定更新

        設定を更新する

        Parameters
        ----------
        key:str
            設定名
        val:str
            設定値
        """
        log = get_logger(__name__)
        data = val
        if self.is_crypt(key) and val is not None:
            data = self.encrypt(val)
            log.info(f"{key} は値を暗号化しました。")

        if key in os.environ:
            os.environ[key] = data
        else:
            self._cfg[key] = data

        # 「_dir」を確認する
        if key.casefold().endswith("_dir"):
            if data is not None and not os.path.isdir(data):
                log.warning(f"'{key}' のディレクトリは存在していません。")

    def delete_config(self, key: str):
        """設定削除

        設定を削除する

        Parameters
        ----------
        key:str
            設定名
        user_id:str
            データベース更新時のユーザーID
        """
        pos = self.has_config(key)
        if pos == 1:
            os.environ.pop(key)
        elif pos == 2:
            del self._cfg[key]

    def has_config(self, key: str) -> int:
        result: int = 0
        if key in os.environ:
            result = 1
        elif key in self._cfg:
            result = 2

        return result

    def is_crypt(self, key: str) -> bool:
        # 暗号化対象の項目か確認する
        items = os.environ.get("hash_key", "database").split(",")
        return key in items

    def encrypt(self, word: str) -> str:
        """文字列暗号化

        文字列を暗号化する。

        Parameters
        ----------
        word:str
            暗号化したい文字列

        Returns
        -------
        str
            暗号化した文字列
        """

        # check word len
        if word is None:
            return None
        elif len(word) <= 0:
            return ""
        elif not self.__init_key:
            return word

        f = Fernet(self.get_config(KEY_NAME).encode())
        return f.encrypt(word.encode()).decode()

    def decrypt(self, word: str) -> str:
        """文字列復号化

        暗号化した文字列を復号する。

        Parameters
        ----------
        word:str
            暗号化した文字列

        Returns
        -------
        str
            復号した文字列。復号できない場合は渡された文字列
        """
        if word is None:
            return None
        elif len(word) <= 0:
            return ""
        elif not self.__init_key:
            return word

        f = Fernet(self.get_config(KEY_NAME).encode())
        try:
            msg = f.decrypt(word.encode()).decode()
        except Exception:
            get_logger(__name__).warning("文字列を復号出来ませんでした(非暗号化？)。")
            msg = word

        return msg

    def use_ssl(self) -> bool:
        use_ssl: str = self.get_config("use_ssl", "0")
        return use_ssl == "1"
