from sqlalchemy import select

from app import get_logger, SYSTEM_USER_ID
from app.db import Engine, DbSession, get_ses_maker
from app.crud.configure import get_all
from app.models import Base
from app.models.configure import Configure
from app.service.configure import SayTorConfig
from app.service.users import check_sys_users


async def get_version() -> str:
    """システムバージョン取得

    設定からシステムバージョンを取得する

    Returns
    -------
    str
        システムバージョン。取得に失敗した場合、空文字
    """
    log = get_logger(__name__)
    db_version = ""
    try:
        async with DbSession() as ses:
            res = await ses.scalars(select(Configure).where(Configure.ckey == "sysVersion"))
            cfg = res.first()
            if cfg is not None:
                db_version = cfg.cval
            else:
                log.debug("システムバージョンがデータベース中に見つかりませんでした。")
    except Exception:
        log.warn("データベースが未設定と判断しました。")

    return db_version


async def init_db() -> bool:
    """データベース初期化

    使用するデータベースの初期化を行う。

    Return
    ------
    bool
        正常終了時、True
    """
    # https://stackoverflow.com/questions/20744277/sqlalchemy-create-all-does-not-create-tables
    # You should put your model class before create_all() call, like this:
    import app.models.mail_flow

    ret_code: bool = False
    log = get_logger(__name__)

    row = Configure()
    row.ckey = "sysVersion"
    row.cval = SayTorConfig().get_config("sysVersion", "1.0.0")
    row.update_by = SYSTEM_USER_ID

    try:
        async with Engine.begin() as conn:
            Base.metadata.bind = Engine
            await conn.run_sync(Base.metadata.create_all)

        async with DbSession() as ses:
            ses.add(row)
            await ses.commit()

        # システムユーザーの作成
        ret_code = await check_sys_users()
    except Exception:
        log.warning("データベースの初期化処理でエラーが発生しました。", exc_info=True)

    return ret_code


async def has_activation() -> bool:
    from app.service.license import ACTIVATION_KEY, create_trial

    """起動時のアクティベーションコード確認
    
    設定されたアクティベーションコードを確認し、無ければ作成する。

    Return
    ------
    bool
        正常終了時、True
    """
    log = get_logger(__name__)
    cfg = SayTorConfig()

    # アクティベーションコードがあればそのまま終了
    act: str = cfg.get_config(ACTIVATION_KEY, "")
    if len(act) > 0:
        log.info(f"アクティベーションコードが見つかりました。= {act}")
        return True
    else:
        log.warning("アクティベーションコードが見つからないため、試用版アクティベーションコードを発行します。")

    # 試用版ライセンスコード作成
    code: str = create_trial()
    if len(code) <= 0:
        log.error("試用版アクティベーションコードが発行できませんでした。")
        return False

    row = Configure()
    row.ckey = ACTIVATION_KEY
    row.cval = code
    row.update_by = SYSTEM_USER_ID
    ret_code: bool = False

    try:
        async with DbSession() as ses:
            ses.add(row)
            await ses.commit()

        # 作成したコードを設定
        cfg.set_config(ACTIVATION_KEY, code)
        ret_code = True
        log.info("試用版のアクティベーションコードを設定しました。")
    except Exception:
        log.exception("試用版のアクティベーションコードが設定できませんでした。")

    return ret_code


async def load_config():
    """データベース設定取得

    データベースから設定を取得し、設定に反映する

    """
    from app.service.configure import SayTorConfig

    cache = {}

    async with get_ses_maker() as db:
        lst: list[Configure] = await get_all(db)

    for i in lst:
        cache[i.ckey] = i.cval
    SayTorConfig()._cfg.update(cache)
