import os

# from memory_profiler import profile

from app import get_logger
from app.db import get_ses_maker
from app.crud.mail_flow import create, get_count
from app.enum import FlowTerm
from app.service.action import get_new_action
from app.service.recv_data import ReadRecvData
from app.service.send_eml import SendMailMsg
from app.service.filter_cache import MailFilterCache
from app.service.mail_send import new_sender
from app.service.configure import SayTorConfig
from app.util import json_escape_char


class ApplyFilter:
    def __init__(self, filepath: str):
        self.file = ReadRecvData(filepath)
        info = os.stat(filepath)
        self.size = info.st_size

    # @profile(precision=4)
    async def check_and_forward(self):
        flt = None
        send_mail = None
        flow_match_id: list[int] = []  # マッチしたフィルターのID(流量登録用)
        base_name: str = os.path.basename(self.file.file_name)
        log = get_logger(__name__)

        # サイズ規定
        mx_str: str = SayTorConfig().get_config("accept_max_size", "0")
        if mx_str.isnumeric():
            mx_size: int = int(mx_str)
            if mx_size > 0 and self.size > mx_size:
                log.warning(f"メールファイル {base_name} のサイズが大きすぎます。強制的にフィルターされました。上限={mx_size:,} メールサイズ={self.size:,}")
                return
        else:
            log.debug("メールサイズの上限設定は無視されます。")

        log.info(f"{base_name} へのメールフィルター適用開始。")

        # フィルター適用ループ
        for f in MailFilterCache().get():
            log.debug(f"メールフィルター ID: {f.filter_id} のフィルターチェック開始。")

            # フィルター有効確認
            if not f.is_executable():
                log.debug(f"メールフィルター ID: {f.filter_id} は現在無効となっています。")
                continue

            # フィルター適用確認
            if not f.valid(self.file):
                continue

            # フィルター判定を通ったので、流量として登録予定
            # 流量条件があるので、リストにする
            flow_match_id.append(f.filter_id)

            # 流量チェック
            if f.enable_flow:
                log.detail(f"メールフィルター ID: {f.filter_id} 流量有効。条件:{f.flow_term}")
                # 件数取得
                async with get_ses_maker() as db:
                    cnt = await get_count(db, filter_id=f.filter_id, term_minute=f.flow_interval)

                if f.flow_term == FlowTerm.greater_equ:
                    # より大きい指定で以下だったのでスキップ
                    if f.flow_count is not None and f.flow_count > cnt:
                        log.info(f"メールフィルター ID: {f.filter_id} 流量(gt)不一致のため適用除外 設定数:{f.flow_count} > 流量:{cnt}")
                        continue
                    else:
                        log.detail(f"メールフィルター ID: {f.filter_id} 流量(gt)一致。フィルター適用 設定数:{f.flow_count} <= 流量:{cnt}")
                else:
                    # より小さい指定で以上だったのでスキップ
                    if f.flow_count is not None and f.flow_count <= cnt:
                        log.info(f"メールフィルター ID: {f.filter_id} 流量(lt)不一致のため適用除外 設定数:{f.flow_count} <= 流量:{cnt}")
                        continue
                    else:
                        log.detail(f"メールフィルター ID: {f.filter_id} 流量(lt)一致。フィルター適用 設定数:{f.flow_count} > 流量:{cnt}")
            else:
                log.detail(f"メールフィルター ID: {f.filter_id} 流量無効。")

            # 該当フィルター発見、フィルターチェック終了
            flt = f
            break

        log.debug("メールフィルターの適用終了。")

        if flt:
            log.debug(f"メールフィルター ID: {flt.filter_id} のログフィルター適用開始。")

            # ログフィルター確認
            send_mail = flt.check_mail_attach(self.file)
        else:
            log.debug(f"メールフィルターの対象外だったため、メール {base_name} は転送されます。")

            # 改行補正(これを入れないと送信メールが改行されない)
            body_data: str = self.file.get_body()
            if "\r" in body_data:
                body_data = body_data.replace("\r\n", "\n")

            # 該当フィルター無し。強制送信
            send_mail = SendMailMsg(
                subject=self.file.get_subject(),
                body=body_data,
                attach_file=self.file.store_all_attach(),
            )

        # 転送確認
        if send_mail is not None:
            # メール送信サーバーインスタンスの取得
            smtp = await new_sender()
            if smtp:
                # 送信を実施して、結果をどちらかに保存
                if smtp.send(send_mail):
                    # 送信済み
                    send_mail.store_sended()
                else:
                    # 送信リトライ
                    send_mail.store_sent()
            else:
                # 送信リトライ
                send_mail.store_sent()

        # アクション通知
        if flt:
            # 動的パラメータの準備
            action_variables: dict[str, str] = {}
            action_variables["FilterId"] = str(flt.filter_id)
            action_variables["FilterName"] = flt.filter_name
            action_variables["EmlPath"] = os.path.abspath(self.file.fullpath)
            action_variables["EmlSubject"] = self.file.get_subject()
            action_variables["EmlTo"] = self.file.get_addr_to()
            action_variables["EmlCc"] = self.file.get_addr_cc()
            action_variables["EmlFrom"] = self.file.get_addr_from()
            body_str: str = self.file.get_body()
            action_variables["EmlBody"] = body_str
            action_variables["EncodeBody"] = body_str
            if len(body_str) > 0:
                action_variables["EncodeBody"] = json_escape_char(body_str)

            for act in flt.filter_actions:
                act_sv = get_new_action(act)
                act_sv.set_values(action_variables)
                if await act_sv.execute():
                    log.info(f"メールフィルター ID: {act.filter_id} のアクション種別: {act.act_type} の実行を完了しました。")
                else:
                    log.warning(f"メールフィルター ID: {act.filter_id} のアクション種別: {act.act_type} の実行に失敗しました。")

            action_variables.clear()
            action_variables = None

        # フィルター適用確認
        if len(flow_match_id) <= 0:
            # 適用フィルター無で登録する
            flow_match_id.append(0)

        # 流量登録
        async with get_ses_maker() as db:
            for id in flow_match_id:
                result = await create(db, filter_id=id, mail_name=base_name)

                if result is None:
                    # 流量登録に失敗
                    log.error("メール流量の登録に失敗しましたが、処理を継続します。")
