import sys
import platform
import os
import ctypes

from app import get_logger
from app.service.configure import SayTorConfig
from app.util import join_root_path
from app.db import get_ses_maker
from app.crud.mail_flow import delete_at

# アクティベーションコードキー
ACTIVATION_KEY: str = "activation"

# 製品コード
PRODUCT_CODE: str = "10101"

# イメージキャッシュ
cache_library: ctypes.CDLL = None
fnc_create_trial_key = None
fnc_get_remaining_days = None
fnc_get_max_count = None
fnc_is_trial = None


def load_library():
    # ライブラリのキャッシュ
    global cache_library
    if cache_library is None:
        # バイナリデータ格納ディレクトリ
        bin_dir: str = join_root_path(SayTorConfig().get_config("bin_dir", "binary"))

        # プラットフォーム別のライブラリ名
        if platform.system().casefold() == "Linux".casefold():
            lib_name: str = os.path.join(bin_dir, "AlertLive.so")
        else:
            lib_name: str = os.path.join(bin_dir, "AlertLive.dll")

        # ライブラリが無ければシステム停止
        if not os.path.exists(lib_name):
            get_logger(__name__).error(f"共有ライブラリ {lib_name} が見つかりません。")
            sys.exit(1)

        # ライブラリの読み込み
        cache_library = ctypes.cdll.LoadLibrary(lib_name)

    return cache_library


def create_trial() -> str:
    # API のキャッシュ
    global fnc_create_trial_key
    if fnc_create_trial_key is None:
        fnc_create_trial_key = load_library().CreateTrialKey
        fnc_create_trial_key.restype = ctypes.c_char_p
        fnc_create_trial_key.argtypes = [ctypes.c_char_p]

    # トライアルライセンスを作成
    result: bytes = fnc_create_trial_key(PRODUCT_CODE.encode())
    return result.decode()


def get_remain_day() -> int:
    # 必要なオブジェクトの初期化
    log = get_logger(__name__)
    cfg = SayTorConfig()

    # ファイルパスの準備
    usage_file = os.path.join(join_root_path(cfg.get_config("bin_dir", "binary")), "StartUsage.bin")

    # アクティベーションコードの取得
    act: str = cfg.get_config(ACTIVATION_KEY, "")
    if len(act) <= 0:
        log.error("アクティベーションコードの設定がありません。")
        return -1

    # API のキャッシュ
    global fnc_get_remaining_days
    if fnc_get_remaining_days is None:
        fnc_get_remaining_days = load_library().GetRemainingDays
        fnc_get_remaining_days.restype = ctypes.c_int
        fnc_get_remaining_days.argtypes = [ctypes.c_char_p, ctypes.c_char_p]

    # 日数確認
    lease: int = fnc_get_remaining_days(usage_file.encode(), act.encode())

    return lease


async def check_code():
    # 日数確認
    lease: int = get_remain_day()
    get_logger(__name__).info(f"残りの利用可能日数={lease}")
    if lease <= 0:
        get_logger(__name__).error("試用期限切れ。")
        sys.exit(1)

    async with get_ses_maker() as db:
        await delete_at(db, 3600)


def get_max_filter_count() -> int:
    # アクティベーションの取得
    act: str = SayTorConfig().get_config(ACTIVATION_KEY, "")
    if len(act) <= 0:
        get_logger(__name__).error("アクティベーションコードの設定がありません。")
        sys.exit(1)

    # API のキャッシュ
    global fnc_get_max_count
    if fnc_get_max_count is None:
        fnc_get_max_count = load_library().GetMaxCount
        fnc_get_max_count.restype = ctypes.c_int
        fnc_get_max_count.argtypes = [ctypes.c_char_p]

    # 可能件数を確認する。
    return fnc_get_max_count(act.encode())


def is_trial() -> bool:
    # アクティベーションの取得
    act: str = SayTorConfig().get_config(ACTIVATION_KEY, "")
    if len(act) <= 0:
        get_logger(__name__).error("アクティベーションコードの設定がありません。")
        sys.exit(1)

    # API のキャッシュ
    global fnc_is_trial
    if fnc_is_trial is None:
        fnc_is_trial = load_library().IsTrial
        fnc_is_trial.restype = ctypes.c_int
        fnc_is_trial.argtypes = [ctypes.c_char_p]

    # 製品フラグを確認する。
    return fnc_is_trial(act.encode()) == 1
