import os
import shutil
from copy import deepcopy
from glob import glob
from datetime import timedelta
from smtplib import SMTP, SMTP_SSL
from smtplib import SMTPHeloError, SMTPAuthenticationError, SMTPNotSupportedError, SMTPException

# import objgraph
from tornado.ioloop import PeriodicCallback
from O365 import Account, Message
from requests import HTTPError, RequestException

from app import get_logger
from app.db import get_ses_maker
from app.crud.mail_server import get_as_protocol
from app.enum import ConnProtect, MailProtocol, MailAuthMethod
from app.entity.mail_server import MailServer
from app.service.send_eml import SendMailMsg, eml_to_send_mail
from app.service.base_retry import WithRetry
from app.service.configure import SayTorConfig
from app.util import ob_proc
from app.service.o365_mail import O365MailBase


class BaseSendMail(WithRetry):
    """メール送信ベースクラス

    メールを送信する。

    """

    def __init__(self, svr: MailServer):
        self.svr = deepcopy(svr)

        cfg = SayTorConfig()
        self.output: str = cfg.get_config("sended_dir", "sended")
        if not os.path.exists(self.output):
            os.makedirs(self.output, exist_ok=True)

        # リトライデータ
        super(BaseSendMail, self).__init__(
            r_count=int(cfg.get_config("retry_count", 0)),
            r_wait=int(cfg.get_config("retry_wait", 0)),
            r_exponent=int(cfg.get_config("retry_exponent", 0)),
        )

    def send(self, msg: SendMailMsg):
        """メッセージ送信メソッド

        設定されているメッセージを送信する

        Parameters
        ----------
        msg:SendMailMsg
            送信メッセージ

        Return
        ------
        bool
            送信処理が完了した場合、True
        """
        raise NotImplementedError("'send' is no implement")


class MailSend(BaseSendMail):
    """メール送信クラス

    メールを送信する。

    """

    # タイムアウト
    SMTP_TIMEOUT: float = 10.0

    def connect(self):
        conn = None
        log = get_logger(__name__)

        # 接続処理開始
        loop_count = self.get_loop_count()
        for i in range(1, loop_count + 1):
            try:
                if not self.svr.use_ssl == ConnProtect.ssl:
                    conn = SMTP(self.svr.host, self.svr.port, timeout=MailSend.SMTP_TIMEOUT)
                else:
                    conn = SMTP_SSL(self.svr.host, self.svr.port, timeout=MailSend.SMTP_TIMEOUT)

                if self.svr.use_ssl == ConnProtect.starttls:
                    conn.starttls()
            except Exception as ex:
                log.warning(f"送信メールサーバーへの接続に失敗しました。理由: '{str(ex)}'")
                w_time = self.retry_wait(i)
                log.warning(f"({i}) ... {w_time} ミリ秒後にリトライ")

            if conn:
                log.info(
                    f"送信メールサーバー {self.svr.host}:{self.svr.port}(接続モード:{self.svr.use_ssl}) に接続しました。({i}/{loop_count})"
                )
                break

        return conn

    def auth(self, conn) -> bool:
        log = get_logger(__name__)

        # 認証が無ければ無条件終了
        if not conn:
            log.warning("接続されていないため、認証処理は中断しました。")
            return False
        elif self.svr.auth_method == MailAuthMethod.none:
            log.debug(f"送信メールサーバー {self.svr.host} 認証なし。")
            return True

        log.info(f"送信メールサーバー {self.svr.user_id}@{self.svr.host} 認証処理開始。")
        result: bool = False

        loop_count = self.get_loop_count()
        for i in range(1, loop_count + 1):
            try:
                conn.login(self.svr.user_id, self.svr.password)
                result = True
            except (SMTPHeloError, SMTPAuthenticationError, SMTPNotSupportedError, SMTPException):
                # リトライしないエラー
                log.exception(f"送信メールサーバー {self.svr.user_id}@{self.svr.host} は認証エラーのため、処理を中断します。")
                return False
            except Exception as ex:
                log.warning(f"認証でエラーが発生しました。エラー理由: '{str(ex)}'")
                w_time = self.retry_wait(i)
                log.warning(f"({i}) ... {w_time} ミリ秒後にリトライ")

            if result:
                log.info(f"送信メールサーバー {self.svr.user_id}@{self.svr.host} の認証完了。({i}/{loop_count})")
                break

        return result

    def msg_send(self, conn, msg: SendMailMsg) -> bool:
        # 初期値取得
        log = get_logger(__name__)

        # 基本は全メッセージを送信
        loop_count = self.get_loop_count()

        # 送信処理
        result = True
        for i in range(1, loop_count + 1):
            try:
                conn.send_message(msg.create_eml())
            except Exception as ex:
                result = False
                log.warning(f"メールの送信に失敗しました。メール ID: {msg.id}、エラー理由: '{str(ex)}'")
                w_time = self.retry_wait(i)
                log.warning(f"({i}) ... {w_time} ミリ秒後にリトライ")

            if result:
                log.info(f"メール送信完了。メール ID: {msg.id} ({i}/{loop_count})")
                break

        if not result:
            log.error(f"メール送信に失敗しました。メール ID: {msg.id}")

        return result

    def send(self, msg: SendMailMsg) -> bool:
        # メッセージ無
        if not msg:
            return False

        log = get_logger(__name__)

        # 接続処理実施
        result: bool = False
        conn = self.connect()
        if conn:
            # 認証
            if self.auth(conn):
                # 送信
                result = self.msg_send(conn, msg)

                if result:
                    log.info(f"メール ID: {msg.id} を送信しました。")
                else:
                    log.warning(f"メール ID: {msg.id} は送信されませんでした。")

            # 接続終了
            conn.quit()

        return result


class MsOAuthSend(BaseSendMail, O365MailBase):
    """MS O365 用メール送信クラス

    O365 OAuth2 専用の送信クラス

    """

    def connect(self) -> Account:
        return self.oauth_connect(self.svr)

    def msg_send(self, conn: Account, msg: SendMailMsg) -> bool:
        log = get_logger(__name__)

        result: bool = False
        message: Message = None

        # 新規メッセージの準備
        loop_count: int = self.get_loop_count()
        for i in range(1, loop_count + 1):
            try:
                mailbox = conn.mailbox()
                message = mailbox.new_message()
            except HTTPError as ex:
                log.warning(f"O365 メールボックスの操作に失敗しました。エラー内容: '{str(ex)}'")
                w_time = self.retry_wait(i)
                log.warning(f"({i}) ... {w_time} ミリ秒後にリトライ")
            except RequestException as ex:
                log.warning(f"O365 メールボックスの操作中にエラーが発生しました。処理を中断します。エラー内容: '{ex.response}'")
                return False
            except Exception:
                log.exception("O365 メールボックスの操作でエラーが発生しました。処理を中断します。")
                return False

        if message is None:
            return False

        body_data = msg.body
        if "\r" in body_data:
            body_data = body_data.replace("\r\n", "\n").replace("\r", "")

        # 新規メッセージの設定
        message.subject = msg.subject
        message.to.add(msg.addr_to)
        message.cc.add(msg.addr_cc)
        message.sender.address = msg.addr_from
        message.body = body_data
        message.body_type = "TEXT"  # これが無いと改行されない！
        if len(msg.attach_file) > 0:
            message.attachments.add(msg.attach_file)
        log.debug(f"BODY:{message.body_preview}")

        # 新規メッセージの送信
        result: bool = False
        for i in range(1, loop_count + 1):
            try:
                result = message.send()
                if result:
                    break
            except Exception as ex:
                log.warning(f"O365 メール送信中にエラーが発生しました。エラー理由: '{str(ex)}'")
                w_time = self.retry_wait(i)
                log.warning(f"({i}) ... {w_time} 秒後にリトライ")

        return result

    def send(self, msg: SendMailMsg) -> bool:
        log = get_logger(__name__)
        result: bool = False
        con: Account = self.connect()
        if con:
            result = self.msg_send(con, msg)
            if result:
                log.info(f"メール ID: {msg.id} を送信しました。")
            else:
                log.warning(f"メール ID: {msg.id} は送信されませんでした。")

        return result


async def new_sender() -> BaseSendMail:
    """メール送信インスタンス取得

    設定されているメール送信設定を取得し、対応するインスタンスを返却する。

    Return
    ------
    BaseSendMail
        メール送信インスタンス。設定が無ければ None
    Raise
    -----
    ValueError
        設定に矛盾がある場合
    """
    log = get_logger(__name__)
    async with get_ses_maker() as db:
        svr_lst: list[MailServer] = await get_as_protocol(db, MailProtocol.smtp)

    if len(svr_lst) <= 0:
        log.warning("送信メールの設定がありませんでした。")
        return None

    if svr_lst[0].auth_method == MailAuthMethod.oauth:
        result = MsOAuthSend(svr_lst[0])
    elif svr_lst[0].auth_method in [MailAuthMethod.apop, MailAuthMethod.rpop]:
        msg = f"SMTP サーバーの認証設定ではありません。現在の設定: {svr_lst[0].auth_method.name}"
        log.error(msg)
        raise ValueError(msg)
    else:
        result = MailSend(svr_lst[0])

    svr_lst.clear()
    svr_lst = None

    # objgraph.show_growth()

    return result


async def send_retry():
    """未送信メール再送処理

    未送信フォルダーにファイルがある場合、そのファイルを送信する

    """
    cfg = SayTorConfig()
    log = get_logger(__name__)

    sent_name = cfg.get_config("sent_dir", "sent")
    send_name = cfg.get_config("sended_dir", "sended")
    eml_file = glob(f"{sent_name}/**/*.eml", recursive=True)
    if len(eml_file) <= 0:
        log.info("再送する対象のメールがありませんでした。")
        return

    # 送信準備
    svr = await new_sender()
    if not svr:
        log.warning("送信メールの設定がありませんでした。")
        return

    stop_check_flg: str = os.path.join(cfg.get_config("store_dir", "store"), cfg.get_config("mail_stop", "mail_stop"))
    if os.path.exists(stop_check_flg):
        log.info(f"{stop_check_flg} があるため、メール再送処理をスキップしました。")
        return

    log.info("メールの再送処理を開始します。")

    # ファイル送信
    for f in eml_file:
        # convert eml → SendMailMsg
        m = eml_to_send_mail(f)
        if svr.send(msg=m):
            shutil.move(f, os.path.join(send_name, os.path.basename(f)))


async def stop_retry_send():
    """送信リトライ監視停止

    送信リトライ用ディレクトリの定期監視を停止する。

    """
    log = get_logger(__name__)
    if "SMTP" in ob_proc:
        ob_proc["SMTP"].stop()
        del ob_proc["SMTP"]
        log.info("メールの再送処理を停止しました。")
    else:
        log.warning("再送メール処理がありませんでした。")


async def start_retry_send(smtp_info=None):
    """送信リトライ監視開始

    送信リトライ用ディレクトリの定期監視を開始する。

    Parameters
    ----------
    smtp_info:BaseSendMail
        SMTP設定。未指定時は内部で検索する。
    """
    log = get_logger(__name__)

    smtp = None
    # SMTP 情報の取得
    if smtp_info is None:
        sender = await new_sender()
        if sender is not None:
            smtp = sender.svr
    else:
        smtp = smtp_info

    # ポーリング開始判定
    if smtp is not None:
        log.info(f"間隔: {smtp.poll_minute} 分で、メールの定期再送処理を開始します。")
        ob_proc["SMTP"] = PeriodicCallback(send_retry, timedelta(minutes=smtp.poll_minute))
        ob_proc["SMTP"].start()
    else:
        log.warning("定期再送処理は開始されません。")
