
import os
import json
from datetime import datetime

from O365 import Account, FileSystemTokenBackend

from app import get_logger
from app.entity.mail_server import MailServer

class O365MailBase:
    """O365 メール処理用ベースクラス

    OAuth2 認証の O365 メールのための接続処理まで行うクラス
    """
    def oauth_connect(self, server:MailServer) -> Account:
        log = get_logger(__name__)
        pass_ = server.password

        # ID=クライアントID, PASS=クライアントシークレット
        credentials = (server.user_id, pass_)

        # トークンファイルのリファレンスを取得
        f_path = server.ref_con
        if not os.path.exists(f_path):
            log.error(f"トークンファイル '{f_path}' が見つかりませんでした。")
            return None

        # トークンファイルの内容確認
        try:
            with open(f_path) as fp:
                json.load(fp)
        except Exception as ex:
            log.error(f"トークンファイル '{f_path}' の内容解析に失敗しました。エラー理由: '{str(ex)}'")
            return None
        else:
            self.token_path = os.path.abspath(f_path)
            self.token_dir = os.path.dirname(f_path)
            self.token_file = os.path.basename(f_path)

        # トークンダンプ
        tk_end = FileSystemTokenBackend(token_path=self.token_dir, token_filename=self.token_file)

        # アカウント設定
        account = Account(credentials, token_backend=tk_end)
        need_auth:bool = False

        # 認証の要否チェック
        if not tk_end.check_token():
            log.error("展開後の O365 トークンファイルが見つかりません。")
            return None

        # トークン情報を読み込み
        tk = account.connection.token_backend.load_token()
        if tk is not None:
            # 破棄されている場合、リフレッシュ
            if tk.access_expiration_datetime < datetime.now():
                need_auth = True
                log.debug("指定された O365 トークンが古いため破棄されました。")
            else:
                log.debug("指定された O365 トークンは有効期限内です。")
        else:
            need_auth = True
            log.debug("指定された O365 トークンを読み込めませんでした。")

        # 認証？トークンリフレッシュ？
        if need_auth:
            if not account.connection.refresh_token():
                log.error("O365 トークンのリフレッシュに失敗しました。")
                account = None
            else:
                log.debug("O365 トークンがリフレッシュされました。")

        return account
