
import os
from glob import glob
from datetime import timedelta

from tornado.ioloop import PeriodicCallback

from app import get_logger
from app.service.filter import ApplyFilter
from app.util import ob_proc

FILE_LOCK_TIMEOUT_SEC:float=1.0

class Polling(object):
    def __init__(self, dir_name:str):
        """クラス初期化

        クラスを初期化する。

        Parameters
        ----------
        dir_name:str
            監視対象のディレクトリ
        """
        # ディレクトリチェック
        if not os.path.isdir(dir_name):
            raise ValueError(f"{dir_name} is no directory")

        # ディレクトリ内の現状を取得
        self.observe_dir = dir_name
        self.buffer:list[str] = self.get_files()

    def get_files(self)  -> list[str]:
        """監視対象ディレクトリ内収集

        監視対象ディレクトリ内のファイル・ディレクトリをピックアップする

        Return
        ------
        list[str]   
            監視対象ディレクトリ内のアイテムリスト
        """
        return glob(f"{self.observe_dir}/*")

    def update_diff(self) -> list[str]:
        """新規差分チェック

        保持しているリストとの差分を確認して、新規のファイルに対して処理を開始する

        Return
        ------
        list[str]
            収集された差分
        """
        # 新規に作られたファイルのみをピックアップ
        current:list[str] = self.get_files()
        result:list[str] = list(set(current) - set(self.buffer))
        self.buffer = current

        return result

    async def recv_check(self):
        """新規差分チェック

        保持しているリストとの差分を確認して、新規のファイルに対して処理を開始する

        """
        # 新規に作られたファイルのみをピックアップ
        result:list[str] = self.update_diff()

        # 新規の差分に対する処理を開始
        for f in result:
            await ApplyFilter(f).check_and_forward()

def add_rcv_dir(dir_path:str):
    """受信データディレクトリ追加

    監視対象ディレクトリを追加する

    Parameters
    ----------
    dir_path:str
        監視対象のディレクトリ
    """
    if dir_path in ob_proc.keys():
        ob_proc[dir_path].stop()
    
    o = Polling(dir_path)
    ob_proc[dir_path] = PeriodicCallback(o.recv_check, timedelta(seconds=1))
    ob_proc[dir_path].start()
    get_logger(__name__).info(f"'{dir_path}' を受信ファイル格納ディレクトリとして追加しました。監視を開始します。")

def add_sent_dir(dir_path:str):
    """監視対象ディレクトリ追加

    監視対象ディレクトリを追加する

    Parameters
    ----------
    dir_path:str
        監視対象のディレクトリ
    """
    if dir_path in ob_proc.keys():
        ob_proc[dir_path].stop()
    
    o = Polling(dir_path)
    ob_proc[dir_path] = PeriodicCallback(o.recv_check, timedelta(seconds=1))
    ob_proc[dir_path].start()
    get_logger(__name__).info(f"'{dir_path}' を再送メールファイル格納ディレクトリとして追加しました。監視を開始します。")
