import os
import poplib
import ssl
import uuid
import shutil
from copy import deepcopy
from datetime import timedelta

from O365 import Account
from requests import HTTPError, RequestException
from tornado.ioloop import PeriodicCallback

# from memory_profiler import profile

from app import get_logger
from app.enum import ConnProtect, MailProtocol, MailAuthMethod
from app.entity.mail_server import MailServer
from app.service.configure import SayTorConfig
from app.util import ml_proc
from app.util.file_util import get_temp_dir, peaceably_rd
from app.service.base_retry import WithRetry
from app.service.o365_mail import O365MailBase

ML_SERVER_KEY = lambda host, port: f"{host}:{port}"


class BaseRecv(WithRetry):
    """受信メール保存ベースクラス

    受信したメールデータを所定の位置にファイルとして保存するベース。

    """

    def __init__(self, info):
        self.info: MailServer = deepcopy(info)
        if self.info.protocol != MailProtocol.pop3:
            raise ValueError("require pop3 protocol")

        cfg = SayTorConfig()
        self.output: str = cfg.get_config("recv_dir", "data")
        if not os.path.exists(self.output):
            os.makedirs(self.output, exist_ok=True)

        self.stop_check_flg: str = os.path.join(
            cfg.get_config("store_dir", "store"), cfg.get_config("mail_stop", "mail_stop")
        )

        # リトライデータ
        super().__init__(
            cfg.get_config("retry_count", 0),
            cfg.get_config("retry_wait", 0),
            cfg.get_config("retry_exponent", 0),
        )

    def new_file(self) -> tuple[str]:
        """メールファイルパス生成

        所定の場所への出力パスを生成する。
        パスだけ生成する。

        Return
        ------
        tuple[str]
            生成したフルパス, 生成したファイル名
        """
        fname = f"{uuid.uuid4().hex}.eml"
        fpath = os.path.join(self.output, fname)
        return (fpath, fname)

    def check_skip(self) -> str:
        """メール受信スキップ

        スキップフラグファイルがあるかどうかを確認する。

        Return
        ------
        str
            スキップフラグファイルが存在する場合、ファイルパスを返却する
        """
        cfg = SayTorConfig()
        flag_file: str = os.path.join(cfg.get_config("store_dir", "store"), cfg.get_config("mail_stop", "mail_stop"))

        if os.path.exists(flag_file):
            return flag_file

        return ""

    def mail_check(self):
        """メールチェック処理

        指定サーバーからメールを受信する

        """
        raise NotImplementedError("'mail_check not implement")


class PopStore(BaseRecv):
    """受信メール保存クラス

    受信したメールデータを所定の位置にファイルとして保存する。

    """

    # タイムアウト
    POP_TIMEOUT: float = 10.0

    def mail_check(self):
        """メールチェック処理

        指定サーバーからメールを受信する

        """
        flag_check: str = self.check_skip()
        if len(flag_check) > 0:
            get_logger(__name__).info(f"{flag_check} があるため、{self.info.host}:{self.info.port} のメール受信をスキップしました。")
            return

        # 接続
        client = self.connect()
        if client is not None:
            # 認証
            if self.auth(client):
                # 受信
                self.recv(client)

            # セッションクローズ
            client.quit()
            client = None

    def connect(self):
        """サーバー接続処理

        POP サーバーとの接続を開始する

        Returns
        -------
        POP3, POP33_SSL
            接続情報。接続失敗時は None
        """
        log = get_logger(__name__)
        client = None
        ctx = None

        # タイムアウト値取得
        pop_timeout: float = -1
        cfg = SayTorConfig()
        cfg_val: str = cfg.get_config("pop_conn_time", "")
        try:
            if len(cfg_val) > 0:
                pop_timeout = float(cfg_val)
        except Exception:
            log.warning(f"設定されたタイムアウト値が '{cfg_val}' でした。")
        if pop_timeout <= 0:
            pop_timeout = PopStore.POP_TIMEOUT
        log.debug(f"POP3 タイムアウトは {pop_timeout} 秒で設定")

        # SSL 対応時のコンテキストを作成
        if self.info.use_ssl == ConnProtect.starttls or self.info.use_ssl == ConnProtect.ssl:
            ctx = ssl.create_default_context()

        # リトライチェック
        loop_count: int = self.get_loop_count()
        for i in range(1, loop_count + 1):
            try:
                log.info(f"メールサーバー {self.info.host} 接続開始。({i})")

                if self.info.use_ssl == ConnProtect.starttls:
                    client = poplib.POP3(self.info.host, self.info.port, timeout=pop_timeout)
                    client.stls(ctx)
                elif self.info.use_ssl == ConnProtect.ssl:
                    client = poplib.POP3_SSL(self.info.host, self.info.port, timeout=pop_timeout, context=ctx)
                else:
                    client = poplib.POP3(self.info.host, self.info.port, timeout=pop_timeout)

                # POP ログ出力
                debug_level: int = 0
                cfg_val = cfg.get_config("pop_debug", "0")
                if cfg_val.isnumeric():
                    debug_level = int(cfg_val)
                client.set_debuglevel(0 if debug_level < 0 else debug_level)
            except Exception as ex:
                log.warning(f"メールサーバーへの接続に失敗。エラー理由: '{ex}'")
                w_time = self.retry_wait(i)
                log.warning(f"({i}) ... {w_time} ミリ秒後にリトライ")

            if client is not None:
                log.info(f"メールサーバー {self.info.host} への接続完了。({i}/{loop_count})")
                break

        return client

    def auth(self, client) -> bool:
        """ユーザー認証

        POP3 のユーザー認証を行う

        Parameters
        ----------
        client: POP3
            接続情報

        Returns
        -------
        bool
            認証が完了した場合、True

        """
        log = get_logger(__name__)

        # 接続開始
        result: bool = False

        # リトライチェック
        loop_count: int = self.get_loop_count()
        for i in range(1, loop_count + 1):
            try:
                log.info(f"メールサーバー {self.info.user_id}@{self.info.host} 認証開始。({i})")
                pw = self.info.password

                if self.info.auth_method == MailAuthMethod.apop:
                    # APOP 認証
                    client.apop(self.info.user_id, pw)
                elif self.info.auth_method == MailAuthMethod.rpop:
                    # RPOP 認証
                    client.rpop(self.info.user_id)
                    client.pass_(pw)
                else:
                    # ベーシック認証
                    client.user(self.info.user_id)
                    client.pass_(pw)
                result = True
            except Exception as ex:
                log.warning(f"メールサーバーとの認証に失敗しました。エラー理由: '{str(ex)}'")
                w_time = self.retry_wait(i)
                log.warning(f"({i}) ... {w_time} ミリ秒後にリトライ")

            if result:
                log.info(f"メールサーバー {self.info.user_id}@{self.info.host} 認証完了。({i}/{loop_count})")
                break

        return result

    # @profile(precision=4)
    def recv(self, client) -> int:
        """メール受信

        メールを受信して、所定のディレクトリにファイル出力する
        保存したメールはサーバーから削除する

        Parameters
        ----------
        client: POP3
            接続情報

        Returns
        -------
        int
            保存メール数(not 受信メール数)
        """
        log = get_logger(__name__)
        m_count = 0

        # ステータス情報取得
        msg_num = client.stat()[0]
        if msg_num <= 0:
            log.info("メールサーバーにメールがありません。")
            return 0

        tmp_dir: str = get_temp_dir()

        # メール受信ループ
        for i in range(1, msg_num + 1):
            try:
                # 一時ファイルの生成
                fpath = self.new_file()
                tmp_path: str = os.path.join(tmp_dir, fpath[1])
                log.detail(f"一時受信ファイル {tmp_path}")

                # 一時ファイルに保存
                sz = 0
                with open(tmp_path, "wb") as wp:
                    log.debug(f"メール {i} 受信開始。")
                    for line in client.retr(i)[1]:
                        sz += wp.write(line)
                        sz += wp.write(b"\n")
                    log.debug(f"メール {i} 受信完了。")

                # ファイルを受信フォルダーに移動
                shutil.move(tmp_path, fpath[0])
                log.info(f"メール受信を受信しました。HOST:{client.host}, FILE:{fpath[1]}, SIZE:{sz}")
                m_count += 1
            except Exception:
                log.exception(f"({i}/{msg_num}) のメール受信に失敗しました。このメールの受信を一旦スキップします。")
            else:
                # 正常受信後、メール削除
                client.dele(i)

        # ディレクトリの削除
        peaceably_rd(tmp_dir)

        return m_count


class O365Store(BaseRecv, O365MailBase):
    def connect(self) -> Account:
        return self.oauth_connect(self.info)

    def recv(self, conn: Account) -> int:
        log = get_logger(__name__)
        inbox = None

        # メールボックスの取得
        loop_count: int = self.get_loop_count()
        for i in range(1, loop_count + 1):
            try:
                mailbox = conn.mailbox()
                inbox = mailbox.inbox_folder()
            except HTTPError as ex:
                log.warning(f"O365 メールボックスの操作に失敗しました。エラー情報: '{str(ex)}'")
                w_time = self.retry_wait(i)
                log.warning(f"({i}) ... {w_time} ミリ秒後にリトライ")
            except RequestException as ex:
                log.warning(f"O365 メールボックスの操作中にエラーが発生しました。処理を中断します。エラー内容: '{ex.response}'")
                return -2
            except Exception:
                log.exception("O365 メールボックスの操作でエラーが発生しました。処理を中断します。")
                return -2

        if inbox is None:
            return -2

        # メールの取得
        loop_count: int = self.get_loop_count()
        for i in range(1, loop_count + 1):
            try:
                for message in inbox.get_messages(download_attachments=True):
                    fpath = self.new_file()
                    if not message.save_as_eml(to_path=fpath[0]):
                        log.warning("受信メールの保存に失敗しました。一旦処理をスキップしました。")
                    elif not message.delete():
                        log.error("受信メールをサーバーから削除できませんでした。")
                    else:
                        log.info(f"メールを受信して {fpath[1]} として保存しました。")
            except Exception as ex:
                log.warning(f"O365 メール受信中にエラーが発生しました。エラー理由: '{str(ex)}'")
                w_time = self.retry_wait(i)
                log.warning(f"({i}) ... {w_time} 秒後にリトライ")

    def mail_check(self):
        """メールチェック処理

        指定サーバーからメールを受信する

        """
        # 接続
        flag_check: str = self.check_skip()
        if len(flag_check) > 0:
            get_logger(__name__).info(f"{flag_check} があるため、{self.info.host}:{self.info.port} のメール受信をスキップしました。")
            return

        client = self.connect()
        if client is not None:
            # 受信
            self.recv(client)


async def has_server(host: str, port: int) -> bool:
    """メールサービス起動確認

    ポーリング中のメールサービスに当該のサーバーが存在するか確認する

    Parameters
    ----------
    host:str
        ホスト名
    port:int
        ポート番号

    Return
    ------
    bool
        存在している場合、True
    """
    key = ML_SERVER_KEY(host, port)
    return key in ml_proc.keys()


async def stop_server(host: str, port: int) -> bool:
    """メールサービス停止

    ポーリング中のメールサービスを停止する

    Parameters
    ----------
    host:str
        ホスト名
    port:int
        ポート番号

    Return
    ------
    bool
        停止した場合、True
    """
    if not await has_server(host, port):
        return False

    key = ML_SERVER_KEY(host, port)
    ml_proc[key].stop()
    del ml_proc[key]

    return True


async def start_server(svr: MailServer, is_exist: bool = False) -> bool:
    """メールサービス開始

    ポーリング中のメールサービスを開始する

    Parameters
    ----------
    svr: MailServer
        メールサーバー情報
    is_exist:bool
        既存の場合、開始扱いとするなら True

    Return
    ------
    bool
        開始した場合、True
    """
    if await has_server(svr.host, svr.port):
        return is_exist

    # メールサーバーの準備とポーリング開始
    key = ML_SERVER_KEY(svr.host, svr.port)
    if svr.auth_method == MailAuthMethod.oauth:
        pop = O365Store(svr)
    else:
        pop = PopStore(svr)
    ml_proc[key] = PeriodicCallback(pop.mail_check, timedelta(minutes=svr.poll_minute))
    ml_proc[key].start()

    return True


async def start_all_pop3():
    from app.db import get_ses_maker
    from app.crud.mail_server import get_as_protocol

    log = get_logger(__name__)

    # 全てのメールサーバー取得
    async with get_ses_maker() as db:
        servers = await get_as_protocol(db, MailProtocol.pop3)
        for s in servers:
            if await start_server(s, is_exist=True):
                log.info(f"メールサーバー {s.host}:{s.port} の定期受信を開始しました。")
            else:
                log.error(f"メールサーバー {s.host}:{s.port} の定期受信が開始出来ませんでした。")
