import os
import pathlib
from time import sleep
from uuid import uuid4
from email.message import Message
from email.parser import BytesParser
from email.header import decode_header

from app import get_logger
from app.util.archive import BomCab
from app.service.attach_file import AttachFiles
from app.util.file_util import get_temp_dir, peaceably_rd


class BaseReadMail:
    def __init__(self, filepath: str):
        self.fullpath = filepath

    """メールデータ読み取りベースクラス

    メールデータを読み取るためのベースクラス

    """

    def get_subject(self) -> str:
        """タイトル取得

        メールタイトルを取得する

        Return
        ------
        str
            メールタイトル
        """
        raise NotImplementedError("no implement 'get_subject'")

    def store_all_attach(self) -> list[str]:
        """全添付ファイル保存

        添付ファイルをすべて保存する

        Return
        ------
        list[str]
            保存した添付ファイルのパス
        """
        raise NotImplementedError("no implement 'store_all_attach'")

    def get_addr_to(self) -> str:
        """送信先アドレス取得

        メールで TO に指定された値を取得する

        Return
        ------
        str
            TO 文字列
        """
        raise NotImplementedError("no implement 'get_addr_to'")

    def get_addr_from(self) -> str:
        """送信元アドレス取得

        メールで FROM に指定された値を取得する

        Return
        ------
        str
            FROM 文字列
        """
        raise NotImplementedError("no implement 'get_addr_from'")

    def get_addr_cc(self) -> str:
        """CCアドレス取得

        メールで CC に指定された値を取得する

        Return
        ------
        str
            CC 文字列
        """
        raise NotImplementedError("no implement 'get_addr_cc'")

    def get_body(self) -> str:
        """本文取得

        メール本文を取得する

        Return
        ------
        str
            本文
        """
        raise NotImplementedError("no implement 'get_body'")


class ReadMailFile(BaseReadMail):
    """eml ファイル処理クラス

    eml 保存されたメールファイルを解析・処理するクラス

    """

    ENCODINGS = ["iso-2022-jp", "utf-8"]  # デフォルトエンコード

    def __init__(self, filepath: str):
        """初期処理

        eml ファイルからメール情報を取得する

        Parameters
        ----------
        filepath:str
            eml ファイル
        """
        log = get_logger(__name__)
        par = BytesParser()
        self.file_name = os.path.basename(filepath)
        retry_cnt: int = 3
        self.msg: Message = None
        while retry_cnt > 0:
            try:
                self.msg = par.parsebytes(pathlib.Path(filepath).read_bytes())
                retry_cnt = 0
            except PermissionError:
                if retry_cnt > 0:
                    log.warning(f"ファイル {self.file_name} の処理待ちです。{retry_cnt}")
                    retry_cnt -= 1
                    sleep(0.3)
            except Exception:
                log.exception(f"ファイル {self.file_name} の処理エラーです。")
                retry_cnt = 0

        if not self.msg:
            raise ValueError(f"ファイル {self.file_name} の処理に失敗しました。")
        super().__init__(filepath=filepath)

    def __decode_str(self, payload):
        """文字列デコード処理

        メールパーツをデコードした形式で返却する

        Parameters
        ----------
        payload
            デコードする対象

        Return
        ------
        str
            デコードした結果を半角スペースで結合した結果

        """
        if not payload:
            return ""

        try:
            dec = decode_header(payload)
            addr = []
            for i in dec:
                if i[1]:
                    addr.append(i[0].decode(i[1]))
                else:
                    addr.append(i[0].decode())
            return " ".join(addr)
        except Exception:
            return payload

    def get_subject(self) -> str:
        """タイトル取得

        メールタイトルを取得する

        Return
        ------
        str
            メールタイトル
        """
        return self.__decode_str(self.msg["Subject"])

    def get_addr_to(self) -> str:
        """送信先アドレス取得

        メールで TO に指定された値を取得する

        Return
        ------
        str
            TO 文字列
        """
        return self.__decode_str(self.msg["To"])

    def get_addr_from(self) -> str:
        """送信元アドレス取得

        メールで FROM に指定された値を取得する

        Return
        ------
        str
            FROM 文字列
        """
        return self.__decode_str(self.msg["From"])

    def get_addr_cc(self) -> str:
        """CCアドレス取得

        メールで CC に指定された値を取得する

        Return
        ------
        str
            CC 文字列
        """
        return self.__decode_str(self.msg["Cc"])

    def _try_default_charsets(self, body):
        """本文デコード処理

        キャラクターセットが取得されていない場合、様々なエンコードを試みる

        Parameters
        ----------
        body: bytes
            本文データ

        Return
        ------
        str
            デコード結果
        """
        r = ""
        charsets = self.msg.get_charsets()
        if charsets:
            charsets += ReadMailFile.ENCODINGS
        else:
            charsets = ReadMailFile.ENCODINGS

        for c in charsets:
            try:
                r = body.decode(c)
            except:
                pass
            else:
                break

        if len(r) <= 0:
            r = body.decode()

        return r

    def get_body(self):
        """本文取得

        メール本文を取得する

        Return
        ------
        str
            本文
        """
        charset = self.msg.get_content_charset()
        body = None

        if self.msg.is_multipart():
            part = self.msg.get_payload()[0]
            body = part.get_payload(decode=True)
        else:
            body = self.msg.get_payload(decode=True)

        try:
            if body:
                if charset:
                    return body.decode(charset)
                else:
                    return self._try_default_charsets(body)
            else:
                return ""
        except:
            try:
                return body.decode()
            except Exception:
                return body  # デコードできない場合は生データにフォールバック

    def store_attach(self, store_dir: str, filename: str = "", index: int = 1) -> str:
        """添付ファイル保存

        添付ファイルを保存する

        Parameters
        ----------
        store_dir:str
            保存先のディレクトリ
        filename:str = ""
            保存ファイル名。未指定の場合は添付から取得する
        index:int = 1
            添付ファイルの番号。1 から始まる

        Return
        ------
        str
            保存したパス
        """
        file_path = ""
        base_name = filename

        if self.msg.is_multipart():
            try:
                part: Message = self.msg.get_payload()[index]
                if len(base_name) <= 0:
                    tmp_name = part.get_filename()
                    if not tmp_name:
                        base_name = f"{uuid4().hex}.dmp"
                    else:
                        base_name = self.__decode_str(tmp_name)

                file_path = os.path.join(store_dir, base_name)
                with open(file_path, "wb") as wp:
                    wp.write(part.get_payload(decode=True))
            except Exception:
                file_path = ""

        return file_path

    def store_all_attach(self) -> list[str]:
        """全添付ファイル保存

        添付ファイルをすべて保存する

        Parameters
        ----------
        store_dir:str
            保存先ディレクトリ。未指定時はシステムテンポラリ

        Return
        ------
        list[str]
            保存した添付ファイルのパス
        """
        dir_name = get_temp_dir()

        result: list[str] = []
        i: int = 1
        while True:
            tmp = self.store_attach(dir_name, index=i)
            if len(tmp) <= 0:
                break
            else:
                result.append(tmp)
            i += 1

        if len(result) <= 0:
            peaceably_rd(dir_name)

        return result


class ReadHTTPRecv(BaseReadMail):
    """HTTP 受信サービス CAB ファイル処理クラス

    cab 保存された受信サービスファイルを解析・処理するクラス

    """

    def __init__(self, filepath: str):
        """初期処理

        Cab ファイルからメール情報を取得する

        Parameters
        ----------
        filepath:str
            Cab ファイル
        """
        self.store_dir = get_temp_dir()
        self.file_name = filepath
        cab = BomCab(file_path=filepath)
        if not cab.store(self.store_dir):
            raise RuntimeError("broken cabinet")

        self.cab = cab
        super().__init__(filepath=filepath)

    def store_all_attach(self) -> list[str]:
        at = AttachFiles(self.cab.body_files)
        return at.all_files()

    def get_subject(self) -> str:
        # 看板ファイルの subject データを取得
        return self.cab.subject()

    def get_addr_to(self) -> str:
        return ""

    def get_addr_from(self) -> str:
        return ""

    def get_addr_cc(self) -> str:
        return ""

    def get_body(self) -> str:
        # 看板ファイルの subject 以外のデータを取得して連結
        line: list[str] = self.cab.messages()
        return "\r\n".join(line)


def ReadRecvData(file_path: str) -> BaseReadMail:
    """インスタンス生成

    受信データ読み込みインスタンスを返す

    Parameters
    ----------
    file_path:str
        ファイルのパス

    Return
    ------
    BaseReadMail
        処理に適したファイル
    """
    if file_path.lower().endswith(".cab"):
        return ReadHTTPRecv(file_path)
    else:
        return ReadMailFile(file_path)
