import os
from copy import deepcopy
from email.mime.base import MIMEBase
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from pathlib import Path
from uuid import uuid4

from app import datetime_now_aware, get_logger
from app.service.recv_data import ReadMailFile
from app.service.configure import SayTorConfig
from app.util.file_util import peaceably_clear


class SendMailMsg(object):
    """メールメッセージ処理クラス

    メール送信用の処理を実施するクラス

    """

    ENCODING = "iso-2022-jp"  # デフォルトエンコード

    def __init__(
        self,
        addr_from: str = "",
        addr_to: list[str] = [],
        addr_cc: list[str] = [],
        subject: str = "",
        body: str = "",
        attach_file: list[str] = [],
    ):
        """初期化処理

        初期化処理

        Parameters
        ----------
        addr_from:str=""
            送信元アドレス
        addr_to:list[str]=[]
            送信先アドレス
        addr_cc:list[str]=[]
            送信先CCアドレス
        subject:str=""
            メールタイトル
        body:str=""
            メール本文
        attach_file:list[str]=[]
            添付ファイルパス

        """
        cfg = SayTorConfig()
        self.addr_from = addr_from if len(addr_from) > 0 else cfg.get_config("smtp_addr_from")
        self.subject = subject
        self.body = body
        self.attach_file = deepcopy(attach_file)
        self.id = uuid4().hex
        self.msg_id_form = cfg.get_config("msg_id_rule", "<{}@alert_filter.bom>")

        def copy_addresses(lst: list[str], key_name: str) -> list[str]:
            if len(lst) > 0:
                return deepcopy(lst)
            else:
                adr = cfg.get_config(key_name)
                if len(adr) > 0:
                    return adr.split(",")
            return []

        self.addr_to = copy_addresses(addr_to, "smtp_addr_to")
        self.addr_cc = copy_addresses(addr_cc, "smtp_addr_cc")

    def create_eml(self, skip_verify: bool = False) -> MIMEBase:
        """メール生成処理

        指定されたパラメータでメールを作成する

        Parameters
        ----------
        skip_verify:bool
            検証作業をスキップするかどうか。省略時、False

        Return
        ------
        MIMEBase
            メールメッセージ

        """
        if not skip_verify:
            if len(self.addr_from) <= 0:
                raise ValueError("missing sender address")

            if len(self.addr_to) <= 0 and len(self.addr_cc) <= 0:
                raise ValueError("missing destination address")

        # Check attach files
        att_files: list[Path] = []
        for f in self.attach_file:
            p = Path(f)
            if p.is_file():
                att_files.append(p)

        if len(self.body) <= 0 and len(att_files) <= 0:
            raise ValueError("missing payload")

        body_data: str = self.body
        if "\r" in body_data:
            body_data.replace("\r\n", "\n").replace("\r", "")

        # 添付と本文
        if len(att_files) > 0:
            m = MIMEMultipart()
            m.attach(MIMEText(body_data))  # 本文を最初のパートに
            for f in att_files:
                att = MIMEApplication(f.read_bytes())
                att.add_header("Content-Disposition", "attachment", filename=f.name)
                m.attach(att)
        else:
            m = MIMEText(body_data)

        # メッセージ本体
        m["Subject"] = self.subject
        m["To"] = ",".join(self.addr_to)
        m["From"] = self.addr_from
        m["Cc"] = ",".join(self.addr_cc)
        m["Date"] = datetime_now_aware().isoformat()
        m["Message-ID"] = self.msg_id_form.format(self.id)

        return m

    def __store(self, store_dir: str, skip_verify: bool = False):
        result: bool = False
        log = get_logger(__name__)
        try:
            # ファイル名作成
            full_name = os.path.join(store_dir, f"{self.id}.eml")

            # 保存
            eml = self.create_eml(skip_verify)
            result = SendMailMsg.store_eml(eml, full_name)
            if result:
                log.info(f"メールデータを {full_name} に保存しました。")
                peaceably_clear(self.attach_file)
        except Exception:
            log.exception("メールデータの保存に失敗しました。")

        return result

    def store_sended(self) -> bool:
        store_dir = SayTorConfig().get_config("sended_dir", "sended")
        if not os.path.exists(store_dir):
            os.makedirs(store_dir, exist_ok=True)

        return self.__store(store_dir)

    def store_sent(self) -> bool:
        store_dir = SayTorConfig().get_config("sent_dir", "sent")
        if not os.path.exists(store_dir):
            os.makedirs(store_dir, exist_ok=True)

        return self.__store(store_dir, True)

    @classmethod
    def store_eml(cls, mail: MIMEBase, file_path: str):
        """メッセージファイル保存処理

        メールメッセージを保存する

        Parameters
        ----------
        mail:MIMEBase
            メール本体
        file_path:str
            保存先メールファイルのパス

        Return
        ------
        bool
            保存成功時、True
        """
        result = False

        try:
            with open(file_path, "wb") as wp:
                wp.write(mail.as_bytes())
            result = True
        except Exception:
            get_logger(__name__).exception("メールデータの保存中にエラーが発生しました。")
            result = False

        return result


def eml_to_send_mail(file_path: str) -> SendMailMsg:
    """emlファイル変換"""
    log = get_logger(__name__)
    result: SendMailMsg = None

    fname: str = os.path.basename(file_path)
    cfg = SayTorConfig()
    addr_from: str = cfg.get_config("smtp_addr_from", "")
    addr_to: str = cfg.get_config("smtp_addr_to", "")
    addr_cc: str = cfg.get_config("smtp_addr_cc", "")

    if len(addr_from) <= 0:
        log.warning(f"送信用の 'from' 設定が見つかりません。メールファイル {fname} の送信はキャンセルされました。")
        return None

    if len(addr_to) <= 0 and len(addr_cc) <= 0:
        log.warning(f"送信用の 'to', 'cc' 設定が見つかりません。メールファイル {addr_to} の送信はキャンセルされました。")
        return None

    try:
        eml = ReadMailFile(filepath=file_path)
        lst: list[str] = eml.store_all_attach()

        # from/to/cc は設定から取得する
        result = SendMailMsg(
            addr_from=addr_from,
            addr_to=addr_to.split(","),
            addr_cc=addr_cc.split(","),
            subject=eml.get_subject(),
            body=eml.get_body(),
            attach_file=lst,
        )
    except Exception:
        log.exception("EML ファイルを展開中にエラーが発生しました。")
        result = None

    return result
