import os
import pathlib
from hashlib import sha256

from tornado.ioloop import PeriodicCallback, IOLoop

from app import get_logger

## Global Variables ###########################################################
recv_sign = False  # 割り込み確認フラグ
bg_proc: list[PeriodicCallback] = []  # バックグラウンドプロセス
ob_proc: dict[str, PeriodicCallback] = {}  # ディレクトリ監視
ml_proc: dict[str, PeriodicCallback] = {}  # メール受信


## Functions ##################################################################
def get_root_path() -> str:
    return str(pathlib.Path(__file__).parent.parent.parent)


def join_root_path(*paths: str) -> str:
    return os.path.join(get_root_path(), *paths)


def __exit_proc() -> None:
    global ob_proc
    global bg_proc
    global ml_proc

    log = get_logger(__name__)

    log.debug("バックグラウンドの動作プロセスを停止します。")
    for o in ob_proc.values():
        o.stop()
    ob_proc.clear()
    log.info("ディレクトリのファイル監視を停止しました。")

    for p in bg_proc:
        p.stop()
    bg_proc.clear()
    log.info("再送リトライなどのバックグラウンド処理を停止しました。")

    for m in ml_proc.values():
        m.stop()
    ml_proc.clear()
    log.info("メールの定期受信を停止しました。")

    IOLoop.instance().stop()
    log.info("アラートフィルターを停止しました。")


def signal_handler(sig_num, frame):
    # シグナル受信時、フラグを ON にする
    global recv_sign
    recv_sign = True


def try_exit():
    # フラグが ON の場合、処理停止
    global recv_sign
    if recv_sign:
        __exit_proc()


def at_exit():
    global recv_sign
    if not recv_sign:
        __exit_proc()


## ディレクトリ構成(相対) ###############################################################
BASE_DIR = get_root_path()  # ルートディレクトリ
TMPL_DIR_ROOT = "templates"  # テンプレート格納ディレクトリ
STATIC_DIR_ROOT = "static"  # 静的リソース格納ディレクトリ
STORE_ROOT = os.environ.get("store_dir", "store")  # 可変データ格納ディレクトリ


def get_dict_data(data: dict, keypath: str):
    """連想配列データ取得

    階層付きキーデータ取得。
    / で連結したキーをもとに連想配列からデータを取得する

    Parameters
    ----------
    data: dict
        データを取得する連想配列
    keypath: str
        連想配列のキーを / で連結したもの

    Returns
    -------
    object
        取得されたデータ。該当するキーが無い場合は None
    """
    keys = keypath.split("/")
    if keys[0] == keypath:
        return data[keys[0]] if keys[0] in data else None
    elif keys[0] in data:
        return get_dict_data(data[keys[0]], "/".join(keys[1:]))
    else:
        return None


def convert_password(pass_: str) -> str:
    sum = sha256()
    ans = pass_
    for i in range(10):
        sum.update(ans.encode())
        ans = sum.hexdigest()

    return ans


def parseBool(val: str) -> bool:
    # see https://stackoverflow.com/questions/715417/converting-from-a-string-to-boolean-in-python
    return val.lower() in ["true", "1", "t", "y", "yes", "yeah", "yup", "certainly", "uh-huh"]


def json_escape_char(data: str) -> str:
    """JSON用特殊記号エスケープ

    RFC 8259 に従い、特殊記号をエスケープする

    Parameters
    ----------
    data:str
        対象のデータ

    Return
    ------
    str
        エスケープ後のデータ
    """
    return (
        data.replace('"', '""')
        .replace("\\", "\\\\")
        .replace("/", "\\/")
        .replace("\b", "\\b")
        .replace("\f", "\\f")
        .replace("\n", "\\n")
        .replace("\r", "\\r")
        .replace("\t", "\\t")
    )
