import os
import pathlib
from base64 import b85encode, b85decode
from zipfile import ZipFile, ZIP_DEFLATED
from bz2 import compress, decompress

import yaml
from cabarchive import CabArchive, CabFile
from chardet import detect

# from memory_profiler import profile

from app import get_logger


def zip_str(buff: str) -> str:
    """文字列圧縮

    文字列を圧縮する

    Parameters
    ----------
    buff:str
        圧縮したい文字列

    Return
    ------
    str
        圧縮された文字列
    """
    ret: str = ""
    try:
        bin: bytes = compress(buff.encode())
        ret = b85encode(bin).decode()
    except Exception:
        get_logger(__name__).exception("文字列の圧縮時にエラーが発生しました。")

    return ret


def unzip_str(buff: str) -> str:
    """文字列伸張

    圧縮された文字列を伸張する

    Parameters
    ----------
    buff:str
        圧縮された文字列

    Return
    ------
    str
        伸張された文字列
    """
    ret: str = ""
    try:
        bin: bytes = b85decode(buff)
        ret = decompress(bin).decode()
    except Exception:
        get_logger(__name__).exception("文字列の伸張時にエラーが発生しました。")

    return ret


class AbsBdFile(object):
    """ファイル操作用抽象クラス

    アラートフィルターで処理するファイルをラッピングするクラス

    """

    def __init__(self, file_path: str) -> None:
        """初期化処理。

        クラスの初期化処理

        Parameters
        ----------
        file_path : str
            処理対象のファイルパス。
        """
        self.path: str = file_path
        self.files: list[str] = []

    def store(self, work_dir: str) -> list[str]:
        """ファイル保存

        ファイル保存用の、オーバーライドするメソッド。

        Returns
        -------
        bool
            常に False
        """
        raise NotImplementedError("method STORE not implement")


class BdCabinet(AbsBdFile):
    """キャビネットファイル処理クラス

    キャビネットファイル(.cab)を伸張するためのクラス。

    """

    def __init__(self, file_path: str) -> None:
        """初期化処理

        クラスの初期化処理。

        Parameters
        ----------
        file_path : str
            処理対象のファイルパス
        """
        self.cab: CabArchive = CabArchive(pathlib.Path(file_path).read_bytes())
        super().__init__(file_path)

    def store(self, work_dir: str) -> bool:
        """ファイル保存

        キャビネットファイルの内容をワーキングディレクトリ内にすべて保存する。

        Parameters
        ----------
        work_dir:str
            ワーキングディレクトリ

        Returns
        -------
        bool
            ファイルがすべて保存された場合、True
        """
        if not pathlib.Path(work_dir).is_dir():
            os.makedirs(work_dir, exist_ok=True)

        log = get_logger(__name__)
        self.files.clear()
        result: bool = True
        try:
            for k in self.cab:
                cff: CabFile = self.cab[k]
                fname = os.path.join(work_dir, cff.filename)
                log.debug(f"キャビネット内の {k} を {fname} として伸張します。")

                d_name = str(pathlib.Path(fname).parent)
                os.makedirs(d_name, exist_ok=True)
                with open(fname, "wb") as wp:
                    wp.write(cff.buf)

                self.files.append(fname)
            log.info(f"キャビネットファイル {self.path} の伸張が完了しました。")
        except Exception:
            result = False
            log.exception(f"キャビネットファイル {self.path} の伸張時にエラーが発生しました。")

        return result


class BdZip(AbsBdFile):
    """Zip ファイル処理クラス

    ZIP ファイル(.zip)を処理するためのクラス。

    """

    def __init__(self, file_path: str):
        super().__init__(file_path)
        self.base_encode = ""
        self.jp_encode = ""

    # @profile(precision=4)
    def store(self, work_dir: str) -> bool:
        """ファイル保存

        ZIP ファイルの内容をワーキングディレクトリ内にすべて保存する。

        Returns
        -------
        bool
            すべて正常に保存された場合、True
        """
        result: bool = True
        log = get_logger(__name__)
        self.files.clear()

        try:
            with ZipFile(self.path) as zip:
                # ファイル名のエンコード確認
                for f in zip.filelist:
                    if len(self.base_encode) <= 0:
                        self.base_encode = "utf-8" if f.flag_bits & 0x800 else "cp437"

                    if len(self.jp_encode) <= 0:
                        d = detect(f.orig_filename.encode(self.base_encode))
                        try:
                            if d["encoding"] != "ascii":
                                self.jp_encode = d["encoding"]
                                break
                        finally:
                            d.clear()
                            d = None

                    if len(self.jp_encode) > 0:
                        break

                if len(self.jp_encode) <= 0:
                    self.jp_encode = "ascii"

                # ZIP ファイル展開
                for f in zip.filelist:
                    f.filename = f.orig_filename.encode(self.base_encode).decode(self.jp_encode)

                    if os.sep != "/" and os.sep in f.filename:
                        f.filename = f.filename.replace(os.sep, "/")

                    zip.extract(f, path=work_dir)
                    if not f.is_dir():
                        self.files.append(os.path.join(work_dir, f.filename))

            log.info(f"ZIP ファイル {self.path} の伸長が完了しました。")
        except Exception:
            log.exception(f"ZIP ファイル {self.path} の伸張時にエラーが発生しました。")
            result = False

        return result

    # @profile(precision=4)
    def compress(self) -> bool:
        """Zip ファイル圧縮

        self.files にあるファイルを ZIP ファイルに圧縮する。

        Returns
        -------
        bool
            正常に圧縮された場合、True
        """
        log = get_logger(__name__)
        result: bool = False

        try:
            with ZipFile(self.path, mode="w", compression=ZIP_DEFLATED) as zip:
                for f in self.files:
                    zip.write(f, arcname=os.path.basename(f))
            log.info(f"ファイルを {self.path} に圧縮しました。")
            result = True
        except Exception:
            log.exception(f"{self.path} の作成時にエラーが発生しました。")

        return result


class BomCab(BdCabinet):
    """HTTPS 受信サービス用処理クラス

    Cab を少し拡張した処理クラス。

    """

    def __init__(self, file_path: str):
        """初期化処理

        親クラス初期化後、受信サービス用の値を初期化

        """
        super().__init__(file_path)
        self.message_file = ""
        self.body_files = []

    def store(self, work_dir: str) -> bool:
        """ファイル保存

        CAB ファイルの内容をワーキングディレクトリ内にすべて保存する。
        保存後、看板ファイルと本体ファイルとに分けて管理する。

        Parameters
        ----------
        work_dir:str
            ワーキングディレクトリ

        Returns
        -------
        bool
            すべて正常に保存された場合、True
        """
        result: bool = False

        if super().store(work_dir):
            for f in self.files:
                if os.path.basename(f) == "commandline.yml":
                    self.message_file = f
                else:
                    self.body_files.append(f)

            result = len(self.message_file) > 0 or len(self.body_files) > 0

        return result

    def subject(self) -> str:
        """コマンドタイトル取得

        看板ファイルからコマンドタイトルを取得する。

        Return
        ------
        str
            タイトル
        """
        result: str = ""

        # コマンドラインが無い場合は無条件に不一致
        if len(self.message_file) <= 0:
            return result

        # コマンドラインの中身を確認
        with open(self.message_file, encoding="UTF-8") as fp:
            d = yaml.safe_load(fp)

            # 形式不備
            if "parameters" not in d:
                return result

            for v in d["parameters"]:
                if v["key"] == "s":
                    result = v["value"]
                    break

        return result

    def messages(self) -> list[str]:
        """コマンドメッセージ取得

        看板ファイルからメッセージを取得する。

        Return
        ------
        list[str]
            メッセージ

        """
        result: list[str] = []

        # コマンドラインが無い場合は無条件に不一致
        if len(self.message_file) <= 0:
            return result

        # コマンドラインの中身を確認
        with open(self.message_file, encoding="UTF-8") as fp:
            d = yaml.safe_load(fp)

            # 形式不備
            if "parameters" not in d:
                return result

            for v in d["parameters"]:
                if not v["key"] == "s":
                    result.append(v["value"])

        return result
