
import re
import random
from base64 import b64decode

DECODE_TABLE = [
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, 62, -1, 63,
    52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1,
    -1,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
    15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, 63,
    -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
    41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51                     
]
"""バイトコード判定テーブル
"""

b64ptn = re.compile(r"^[a-z0-9A-Z/\+]+={1,2}$")
"""Base64 判定正規表現
"""

def is_base64(line: str) -> bool:
    """Base64文字列判定

    Base64 でデコードできる文字列か確認する

    Parameters
    ----------
    line : str
        確認対象の文字列

    Returns
    -------
    bool
        デコードできる場合、True
    """

    if len(line) <= 0:
        return False
    
    return is_base64_bytes(line.encode(encoding='utf-8'))

def is_base64_bytes(bin: bytes) -> bool:
    """Base64バイト列判定

    Base64 でデコードできるバイト列か確認する

    Parameters
    ----------
    bin : bytes
        確認対象のバイト列

    Returns
    -------
    bool
        デコードできる場合、True
    """
    if len(bin) <= 0:
        return False

    for b in bin:
        if not is_base64_byte(b) and not is_white_space(b):
            return False

    return True

def is_base64_byte(octet: int) -> bool:
    """Base64文字判定

    Base64 の対象バイトか確認する

    Parameters
    ----------
    octet : int
        確認対象の 1 バイト

    Returns
    -------
    bool
        対象の場合、True
    """
    if octet.to_bytes(1, byteorder='big').decode('utf-8') == '=':
        return True
    if octet >= 0 and octet <= len(DECODE_TABLE) and DECODE_TABLE[octet] != -1:
        return True
    return False

def is_white_space(octet: int) -> bool:
    """ホワイトスペース判定

    ホワイトスペースとなるバイトか確認する

    Parameters
    ----------
    octet : int
        確認対象の 1 バイト

    Returns
    -------
    bool
        対象の場合、True
    """
    ch = octet.to_bytes(1, byteorder='big').decode('utf-8')
    if ch == " " or ch == "\n" or ch == "\r" or ch == "\t":
        return True
    return False

def get_bytes(line: str) -> bytes:
    """文字列→バイトコード変換

    文字列をバイト列に変換する。
    Base64 でデコード出来る場合、Base64 デコードを実施する。

    Parameters
    ----------
    line : str
        対象の文字列

    Returns
    -------
    bytes
        変換されたバイト列
    """
    if is_base64(line):
        return b64decode(line)
    else:
        return line.encode(encoding='utf-8')

def get_mark_char() -> str:
    """
    Base64 で使用しない文字をランダムで取得する。

    Returns
    -------
    str
        ランダムに取得した Base64 で使用しない文字
    """
    i = len(DECODE_TABLE) - 1
    while DECODE_TABLE[i] != -1:
        i = random.randrange(33, len(DECODE_TABLE))

    return chr(i)

def join_mark_char(*words:str) -> str:
    """文字列結合

    Base64 で使用しない文字で文字列結合する。

    Parameters
    ----------
    *word : str
        結合対象の文字列。不定長

    Returns
    -------
    str
        結合された文字列
    """
    if len(words) < 1:
        return ""

    result:str = words[0]
    for w in words[1:]:
        c = get_mark_char()
        result += c + w

    return result

def split_mark_char(line: str) -> list[str]:
    """文字列分割

    Base64 で使用しない文字で文字列分割する。

    Parameters
    ----------
    line : str
        分割対象の文字列

    Returns
    -------
    list[str]
        分割された文字列
    """
    result: list[str] = []
    if len(line) <= 0:
        return []

    buff = ""
    for c in line:
        if is_base64_byte(ord(c)):
            buff += c
        else:
            result.append(buff)
            buff = ""

    if len(buff) > 0:
        result.append(buff)

    return result
