
import os

from app import get_logger
from app.util import parseBool

def get_all_dot_env(file_path:str=".env") -> list[tuple[str,bool]]:
    """.env ファイル読み取り

    環境変数設定をすべて読み込む。
    環境変数とそれ以外を識別するフラグ付きで保存。

    Return
    ------
    list[tuple[str,bool]]
        行データ
    """
    result:list[list[str,bool]] = []
    if not os.path.exists(file_path):
        return result

    with open(file_path, encoding="utf-8") as fp:
        for row in fp:
            line = row.strip()

            if line.startswith("#"):
                val = (line, False)
            elif len(line) <= 0:
                val = (line, False)
            elif line.count("=") > 0:
                val = (line, True)
            else:
                val = (line, False)

            result.append(val)

    return result

def load_env(file_path:str=".env", over_write:bool=None):
    """環境変数ファイル読み込み

    環境変数ファイルを読み込み、環境変数に展開する。

    Parameters
    ----------
    file_path:str
        読み込みファイルのパス。初期値はカレントの .env ファイル
    over_write:bool
        既存の環境変数を更新する場合、True。初期値は False
    """
    if not os.path.exists(file_path):
        return

    over_write_env:bool = parseBool(os.environ.get("overwrite_environ", "False")) if over_write is None else over_write

    try:
        for item in get_all_dot_env(file_path=file_path):
            # 読み取り行の妥当性判定
            if not item[1]:
                continue

            words:list[str] = item[0].split("=", 1)

            # 変数の上書き判定
            if words[0] in os.environ and not over_write_env:
                continue

            val = words[1].strip("\"' ")

            # 環境変数設定
            os.environ[words[0]] = val
    except Exception as ex:
        get_logger(__name__).warning(f"環境変数の設定中にエラーが発生しました。エラー理由: '{str(ex)}'")

def set_env(key:str, value:str, file_path:str=".env", use_quote:bool=True):
    """環境変数ファイル保存

    環境変数ファイルに設定を保存する。

    Parameters
    ----------
    key:str
        環境変数名
    value:str
        設定値
    file_path:str
        読み込みファイルのパス。初期値はカレントの .env ファイル
    use_quote:bool
        値をクォーテーションで括る場合、True
    """
    set_value = f"{key}='{value}'" if use_quote else f"{key}={value}"
    try:
        is_exist:bool = False
        all_lines = get_all_dot_env(file_path=file_path)
        for i in range(len(all_lines)):
            item = all_lines[i]
            # 読み取り行の妥当性判定
            if not item[1]:
                continue

            words:list[str] = item[0].split("=", 1)

            # 変数判定
            if words[0].casefold() == key.casefold():
                all_lines[i] = (set_value, True)
                is_exist = True
                break

        if not is_exist:
            all_lines.append((set_value, True))

        with open(file_path, "w", encoding="utf-8") as wp:
            for item in all_lines:
                wp.write(item[0] + "\n")
    except Exception as ex:
        get_logger(__name__).warning(f"環境変数の永続化中にエラーが発生しました。エラー理由: '{str(ex)}'")

def delete_env(key:str, file_path:str=".env"):
    """環境変数削除結果ファイル保存

    環境変数ファイルの設定を削除する。

    Parameters
    ----------
    key:str
        環境変数名
    file_path:str
        読み込みファイルのパス。初期値はカレントの .env ファイル
    """
    try:
        all_lines = get_all_dot_env(file_path=file_path)
        with open(file_path, "w", encoding="utf-8") as wp:
            for l in all_lines:
                # 読み取り行の妥当性判定
                if l[1]:
                    ckey = l[0].split("=")[0]

                    # 変数判定
                    if ckey.casefold() == key.casefold():
                        continue

                wp.write(l[0] + "\n")
    except Exception as ex:
        get_logger(__name__).warning(f"削除した環境変数設定を保存中にエラーが発生しました。エラー理由: '{str(ex)}'")
