import json
import os
import re
import zipfile
import shutil
from tempfile import gettempdir
from uuid import uuid4

from cabarchive import CabArchive
from chardet import detect

# from memory_profiler import profile

from app import get_logger
from app.enum import FileType
from app.service.configure import SayTorConfig

DEFAULT_LOG_PATTERN: str = r"GRP\d{2}MON\d{2}\..*|bom_archive_ntevent\.csv"


# @profile(precision=4)
def check_file_type(file_path: str) -> FileType:
    """ファイルコンテントチェック

    ファイルタイプを内容で判定する。
    ファイルを実際に通してみて、エラーとならないか確認する。

    Parameters
    ----------
    file_path:str
        処理ファイルのパス

    Return
    ------
    FileType
        ファイルタイプ
    """
    log = get_logger(__name__)
    b_name = os.path.basename(file_path)

    if zipfile.is_zipfile(file_path):
        log.debug(f"{b_name} の識別結果: ZIP")
        return FileType.zip

    with open(file_path, "rb") as fp:
        try:
            CabArchive(fp.read())
        except Exception:
            if fp.seekable():
                fp.seek(0)  # 念のため
        else:
            log.debug(f"{b_name} の識別結果: CAB")
            return FileType.cab

        # エンコード判定
        file_encode: str = ""
        enc: dict = {}
        try:
            enc = detect(fp.read())
            file_encode = enc["encoding"]
        except Exception as ex:
            log.debug(f"{b_name} の内部エンコードの取得でエラーが発生しました。強制的に UTF-8 を指定します。エラー理由: '{str(ex)}'")
            file_encode = "UTF-8"
        finally:
            enc.clear()
            enc = None
            if fp.seekable():
                fp.seek(0)  # 念のため

    # JSON チェックは UTF-8 固定で確認(JSON の仕様なので)
    with open(file_path, "r", encoding="UTF-8") as fp:
        try:
            json.load(fp)
        except Exception as ex:
            log.debug(f"{b_name} の JSON 変換時にエラーが発生しました。エラー理由: '{str(ex)}'")
            if fp.seekable():
                fp.seek(0)  # 念のため
        else:
            log.debug(f"{b_name} の識別結果: JSON")
            return FileType.json
        finally:
            if fp.seekable():
                fp.seek(0)  # 念のため

    # 判定されたエンコードでチェック
    with open(file_path, "r", encoding=file_encode) as fp:
        try:
            # 1 行目が " で始まり , が 15 個以上あれば CSV とする
            line: str = fp.readline()
            if line.count(",") >= 15:
                log.debug(f"{b_name} で BOM の監視結果以上のカンマを検出しました。")
                if line.startswith('"'):
                    log.debug(f"{b_name} の識別結果: BOM CSV")
                    return FileType.csv
                else:
                    log.debug(f"{b_name} の識別結果: BOM4 CSV")
        except Exception as ex:
            # 指定エンコードで読み込めない(=テキストでない)ので、バイナリと判断する
            log.debug(f"{b_name} の識別結果: バイナリ。エラー理由: '{str(ex)}'")
            return FileType.bin

    # プレーンテキスト
    log.debug(f"{b_name} の識別結果: TEXT")
    return FileType.plain


def is_log_file(file_path: str) -> bool:
    """ログファイル判定関数

    ファイル名を特定のパターンで見て、ログファイルかどうかを判定する

    Parameters
    ----------
    file_path:str
        チェック対象のファイルパス

    Return
    ------
    bool
        ログファイルの場合、True
    """
    pattern = SayTorConfig().get_config("log_pattern", DEFAULT_LOG_PATTERN)
    b_name = os.path.basename(file_path)
    return re.match(pattern, b_name, re.I)


def get_temp_dir() -> str:
    """一時ディレクトリ取得

    一時ディレクトリ内にユニークディレクトリを作成する。

    Return
    ------
    str
        ユニークディレクトリのパス名
    """
    # 一時ディレクトリ名の取得
    tmp_root = SayTorConfig().get_config("sys_temp_dir", "")
    if len(tmp_root) <= 0:
        tmp_root = os.path.join(gettempdir(), "alert_filter")

    # そこにユニーク名をつける
    tmp_dir = os.path.join(tmp_root, uuid4().hex)

    # ディレクトリの作成
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir, exist_ok=True)

    # ディレクトリ名の返却
    return tmp_dir


def peaceably_delete(file_path: str) -> bool:
    """ファイル削除処理

    指定ファイルを穏便に削除する。

    Parameters
    ----------
    file_path:str
        対象のファイルパス

    Return
    ------
    bool
        エラーでなければ True
    """
    result: bool = False
    log = get_logger(__name__)

    try:
        if file_path is None:
            log.warning("ファイルパスが指定されていません。")
        elif os.path.exists(file_path):
            os.remove(file_path)
        else:
            log.debug(f"{file_path} が見つかりませんでした。")
        result = True
    except Exception as ex:
        log.warning(f"{file_path} の削除中にエラーが発生しました。エラー理由: '{str(ex)}'")

    return result


def peaceably_rd(dir_path: str, re_create: bool = False) -> bool:
    """ディレクトリ削除処理

    指定ディレクトリを穏便に削除する。

    Parameters
    ----------
    dir_path:str
        対象のパス
    re_create:bool
        削除後にディレクトリを作成する場合、True

    Return
    ------
    bool
        エラーでなければ True
    """
    result: bool = False
    log = get_logger(__name__)

    try:
        if dir_path is None:
            log.warning("ファイルパスが指定されていません。")
        elif os.path.isdir(dir_path):
            shutil.rmtree(dir_path, ignore_errors=True)
        else:
            log.debug(f"{dir_path} が見つかりませんでした。")
        result = True
    except Exception as ex:
        log.warning(f"{dir_path} の削除中にエラーが発生しました。エラー理由: '{str(ex)}'")

    if re_create:
        os.makedirs(dir_path, exist_ok=True)

    return result


def peaceably_clear(file_list: list[str]):
    """ファイルリスト削除処理

    指定ファイルリストとそのディレクトリを穏便に削除する。

    Parameters
    ----------
    file_list:list[str]
        対象のパスリスト
    """
    dirs_list: dict[str, int] = {}

    # 添付ファイルとそのディレクトリを削除する
    for f in file_list:
        peaceably_delete(f)
        dirs_list[os.path.dirname(f)] = 1

    # 使っていたディレクトリも消す
    for d in dirs_list.keys():
        peaceably_rd(d)

    dirs_list.clear()
    dirs_list = None
