import os
import json
from copy import deepcopy

from chardet.universaldetector import UniversalDetector

from app import get_logger
from app.enum import FileType
from app.util.file_util import check_file_type, peaceably_delete


class BaseLogFile(object):
    """基底ログファイル操作クラス

    ログファイル操作の基底クラス

    """

    def __init__(self, file_path: str):
        """初期化処理

        クラスを初期化する

        Parameters
        ----------
        file_path: str
            対象のファイルパス
        """
        self.pos: int = 0
        self.encoding = "SHIFT_JIS"
        self.buffer = []
        self.current = None

        stat = os.stat(file_path)  # FileNotFoundError
        self.size = stat.st_size

        # 文字コードを判定する
        d = UniversalDetector()
        with open(file_path, "rb") as fp:
            for l in fp:
                d.feed(l)
                if d.done:
                    break

        log = get_logger(__name__)
        res = d.close()
        if res:
            self.encoding = res["encoding"]
            log.info("{} の文字コードは {} です。".format(file_path, res["encoding"]))
            res.clear()
            res = None
        else:
            log.info(f"{file_path} の文字コードは分かりませんでした。")
        d = None

        self.path: str = file_path

    def __del__(self):
        """デストラクタ"""
        self.buffer.clear()

    def is_empty(self) -> bool:
        """バッファ状況取得メソッド

        バッファ状態取得メソッド

        Return
        ------
        bool
            バッファが空の場合、True
        """
        return len(self.buffer) <= 0

    def is_eof(self) -> bool:
        """EOF判定メソッド

        読み取りが最後まで行われているか確認する。

        Return
        ------
        bool
            EOF の場合、True
        """
        if self.pos >= self.size:
            return True
        return False

    def keep(self):
        """バッファ保持メソッド

        バッファ保持メソッド
        """
        if self.current is not None:
            self.buffer.append(self.current)

    def get_line(self) -> str:
        """行取得メソッド

        オーバーライド用の行取得メソッド

        """
        raise NotImplementedError("implement GET_LINE method")

    def save(self, **kwarg):
        """バッファ保存メソッド

        オーバーライド用のバッファ保存メソッド

        Parameters
        ----------
        **kwarg
            パラメータ
        """
        raise NotImplementedError("implement SAVE method")


class TextLogFile(BaseLogFile):
    """テキストログファイル操作クラス

    テキストログファイルクラス

    """

    def get_line(self) -> str:
        """行取得メソッド

        ファイルの行データを取得する。

        Return
        ------
        str
            行データ
        """
        result = ""

        if self.is_eof():
            self.current = None
            return ""

        with open(self.path, "r", encoding=self.encoding) as fp:
            fp.seek(self.pos)
            result = fp.readline().rstrip()
            self.pos = fp.tell()

        # 現行バッファをキープ
        self.current = result

        return result

    def save(self, file_path: str = None):
        """バッファ保存メソッド

        バッファのデータを保存する。

        Parameters
        ----------
        file_path:str
            ファイル指定がある場合。未指定の場合は上書き保存される。
        """
        f = file_path if file_path else self.path

        if self.is_empty():
            peaceably_delete(f)
            return

        with open(f, "w", encoding=self.encoding, newline="") as wp:
            wp.write("\r\n".join(self.buffer))
            wp.write("\r\n")


class CsvLogFile(BaseLogFile):
    """CSV ログファイル処理クラス

    本来は CSV 形式のログファイルを処理するクラス。
    ただし、BOM のイベントログが対応不十分(CRLF が混ざる、ダブルクォーテーションをエスケープしない等)
    なので、テキストログと同等の処理を行う方式にする。

    """

    def __init__(self, file_path: str):
        console = get_logger(__name__)
        super().__init__(file_path=file_path)
        self.lines: list[str] = []
        quote_prefix: bool = False

        # 取得した行データをバッファリング
        with open(self.path, "r", encoding=self.encoding, newline=None) as fp:
            for line in fp:
                row = line.rstrip()
                console.detail(f"\t行データ 【{row}】")

                # ダブルクォート含有確認
                ary: list[str] = row.split(",", 15)
                if not quote_prefix and ary[-1].startswith('"'):
                    quote_prefix = True

                if quote_prefix and not row.endswith('"'):
                    for sub_line in fp:
                        sub_row = sub_line.rstrip()
                        row += "\r\n" + sub_row
                        if sub_row.endswith('"'):
                            break
                    console.detail(f"\t複数行判定【{row}】")
                else:
                    console.detail(f"\t単一行判定【{row}】")
                self.lines.append(row)

        self.size = len(self.lines)

    def __del__(self):
        """デストラクタ"""
        super().__del__()
        self.lines.clear()
        self.lines = None

    def get_line(self) -> str:
        """行取得メソッド

        ファイルの行データを取得する。

        Return
        ------
        str
            CSV の行データ（含ヘッダー）
        """
        if self.is_eof():
            self.current = None
            return ""

        # ポイントずらし
        self.current = self.lines[self.pos]
        result: str = self.current
        self.pos += 1
        return result

    def save(self, file_path: str = None):
        """バッファ保存メソッド

        バッファのデータを保存する。

        Parameters
        ----------
        file_path:str
            ファイル指定がある場合。未指定の場合は上書き保存される。
        """
        f = file_path if file_path else self.path

        # ログが空の場合は削除
        if self.is_empty():
            peaceably_delete(f)
            return

        # バッファの保存
        with open(f, "w", encoding=self.encoding, newline="") as wp:
            wp.write("\r\n".join(self.buffer))


class JsonLogFile(BaseLogFile):
    """JSON ログファイル処理クラス

    JSON 形式のログファイルを処理するクラス

    """

    def __init__(self, file_path: str):
        """初期化処理

        クラスを初期化する

        Parameters
        ----------
        file_path: str
            対象のファイルパス
        """
        super().__init__(file_path=file_path)

        # データを文字列として持つ
        data: dict = None
        with open(self.path, "r", encoding=self.encoding) as fp:
            data = json.load(fp)

        # 期待するデータかどうかのフラグを持つ
        self.is_expect_v = True
        self.lines: list[dict[str, any]] = []

        # リスト型のみを対象とする
        if not isinstance(data, list):
            self.is_expect_v = False
            self.lines.append(data)
        else:
            self.lines = deepcopy(data)

        # 行データとして処理
        self.size = len(self.lines)

    def __del__(self):
        super().__del__()
        self.lines.clear()
        self.lines = None

    def get_line(self) -> str:
        """行取得メソッド

        ファイルの行データを取得する。

        Return
        ------
        str
            JSON オブジェクトを文字列にしたもの
        """
        if self.is_eof():
            self.current = None
            return ""

        # バッファ取得
        self.current = self.lines[self.pos]

        # 値のみを文字列にして TAB で連結してみる(キーはフィルター条件外にする)
        # (ブランクで結合するとフィールド間でまたがった文字列にあたるかもしれないので TAB 結合)
        result: str = "\t".join([str(x) for x in self.current.values()])

        self.pos += 1
        return result

    def save(self, file_path: str = None):
        """バッファ保存メソッド

        バッファのデータを保存する。

        Parameters
        ----------
        file_path:str
            ファイル指定がある場合。未指定の場合は上書き保存される。
        """
        f = file_path if file_path is not None else self.path

        if self.is_empty():
            peaceably_delete(f)
            return

        try:
            with open(f, "w", encoding=self.encoding) as wp:
                if self.is_expect_v:
                    json.dump(self.buffer, wp, ensure_ascii=False)
                else:
                    # 1 行分しかデータが無いので、空で無ければそれを保存
                    json.dump(self.buffer[0], wp, ensure_ascii=False)
        except Exception:
            get_logger(__name__).exception(f"JSON ファイル {f} の保存でエラーが発生しました。")


def LogFile(file_path: str) -> BaseLogFile:
    """ログファイル処理インスタンス生成

    ファイルの内容・拡張子に合わせてインスタンスを生成する。

    Parameters
    ----------
    file_path:str
        処理ファイルのパス

    Return
    ------
    BaseLogFile
        ログファイル操作インスタンス。テキストでない場合 None
    """
    log = get_logger(__name__)

    f_type = check_file_type(file_path)
    if f_type == FileType.csv:
        return CsvLogFile(file_path)
    elif f_type == FileType.json:
        return JsonLogFile(file_path)
    elif f_type == FileType.bin:
        return None

    words = os.path.basename(file_path).split(".")
    e_name = ""
    if len(words) > 1:
        e_name = words[-1]

    if e_name.lower() == "csv":
        log.debug("ファイル拡張子:csv")
        return CsvLogFile(file_path)
    elif e_name.lower() == "json":
        log.debug("ファイル拡張子:json")
        return JsonLogFile(file_path)
    else:
        return TextLogFile(file_path)
