import atexit
import os
import signal
import json
import logging.config
import asyncio
import ssl
import sys
from datetime import timedelta
from logging import addLevelName

from tornado.ioloop import PeriodicCallback, IOLoop
from tornado.web import url, Application
from tornado.httpserver import HTTPServer

from app import random_string
from app.router import AlertFilterHeaderTransform
from app.router.login import LoginHandler, AuthLogin, IndexHandler
from app.router.filter import FilterListHandler, FilterItemHandler
from app.router.default import DefaultHandler
from app.router.config import ServerConfigHandler
from app.router.users import UsersHandler, ChangePassHandler
from app.router.mail_server import MailServerHandler
from app.util import join_root_path, TMPL_DIR_ROOT, STATIC_DIR_ROOT
from app.util import signal_handler, try_exit, at_exit
from app.util import bg_proc, ob_proc
from app.util.env_config import load_env
from app.service.configure import SayTorConfig

## システム設定 #####################################################################
load_env()

loop = asyncio.get_event_loop_policy().get_event_loop()

cfg = SayTorConfig()
cfg.set_config("sysVersion", "1.0.0")  # システムバージョンはここで定義

log_conf = join_root_path("log.json")
if os.path.exists(log_conf):
    with open(log_conf) as fp:
        logging.config.dictConfig(json.load(fp))
addLevelName(5, "DETAIL")

# シークレット初期化
cfg.init_key()

## ハンドラーの準備 ###################################################################
hdl = [
    url("/?", IndexHandler, name="index"),
    url("/login/?", LoginHandler, name="login"),
    url("/login/auth/?", AuthLogin, name="auth"),
    url("/filter/list/?", FilterListHandler, name="filterList"),
    url("/filter/item/?", FilterItemHandler, name="filterItem"),
    url(r"/filter/item/(?P<filter_id>\d+)", FilterItemHandler, name="filterItemParam"),
    url("/config/?", ServerConfigHandler, name="serverConfig"),
    url("/config/(\\w+)", ServerConfigHandler, name="serverConfigDelete"),
    url("/mail/config/?", MailServerHandler, name="mailServer"),
    url(r"/mail/config/(?P<host>.+?)/(?P<port>\d+)", MailServerHandler, name="mailServerParam"),
    url("/user/(\\w+)", UsersHandler, name="user"),
    url("/change/user/?", ChangePassHandler, name="showChangePass"),
    url(r"/change/user/(?P<uid>.+)", ChangePassHandler, name="passChange"),
]

## アプリケーションの設定 ################################################################
application = Application(
    hdl,
    template_path=join_root_path(cfg.get_config("template_dir", TMPL_DIR_ROOT)),
    static_path=join_root_path(cfg.get_config("static_dir", STATIC_DIR_ROOT)),
    login_url=cfg.get_config("rootUrl", "/login/auth"),
    debug=cfg.get_config("debugMode", False),
    default_handler_class=DefaultHandler,
    cookie_secret=random_string(),
    transforms=[AlertFilterHeaderTransform],
)

## バックグラウンド処理の定義 ##############################################################
bg_proc.append(PeriodicCallback(try_exit, 100))

## メイン処理始動 ##########################################################################
if __name__ == "__main__":
    from app import get_logger

    log = get_logger(__name__)

    if cfg.has_config("ssl_cert_key") > 0 and cfg.has_config("ssl_secret_key") > 0:
        # SSL 対応
        ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_ctx.load_cert_chain(cfg.get_config("ssl_cert_key"), cfg.get_config("ssl_secret_key"))

        web_svr = HTTPServer(application, ssl_options=ssl_ctx)
        log.info("接続方式=HTTPS")
        cfg.set_config("use_ssl", "1")
    else:
        web_svr = application
        log.info("接続方式=HTTP")

    # シグナルハンドラーの登録(処理中断用)
    signal.signal(signal.SIGINT, signal_handler)

    # 終了ハンドラー
    atexit.register(at_exit)

    # ポート番号取得
    port_val: str = ""
    try:
        port_val = cfg.get_config("port", 8888)
        port_no = int(port_val)
    except ValueError:
        sys.exit(f"ポート番号が正しくありません。指定された値={port_val}")
    else:
        log.debug(f"ポート番号 {port_no} で待機。")

    ## データベース接続 ##*##*##*##*##*##*##*##*##*##*##*##*##*##
    # 設定を生かすため、遅延読み込み
    from app.db import connect

    if not connect():
        sys.exit("データベース接続に失敗しました。")

    from app.service.database import get_version, init_db, has_activation

    gen_db: int = 0  # データベースの新規作成フラグ
    sys_ver = loop.run_until_complete(get_version())
    if len(sys_ver) <= 0:
        if not loop.run_until_complete(init_db()):
            sys.exit("データベースの初期化に失敗しました。")
        else:
            gen_db = 1 << 0
    else:
        log.debug(f"システムバージョン: {sys_ver}")

    if not gen_db:
        from app.service.users import check_sys_users

        if not loop.run_until_complete(check_sys_users()):
            sys.exit("管理者アカウントのチェックに失敗しました。")
        else:
            gen_db = 1 << 1

    from app.service.database import load_config

    # データベース内設定読み込み
    loop.run_until_complete(load_config())  # データベースの設定読み込み

    # アクティベーションコードの確認
    if not loop.run_until_complete(has_activation()):
        sys.exit("アクティベーションコードのチェックに失敗しました。")

    # アクティベーションコードの起動確認と定期確認処理の登録
    from app.service.license import check_code

    loop.run_until_complete(check_code())
    bg_proc.append(PeriodicCallback(check_code, timedelta(hours=24)))
    log.info("起動中は 24 時間ごとにアクティベーションコードがチェックされます。")

    from app.service.filter_cache import MailFilterCache

    # フィルターのキャッシュ
    loop.run_until_complete(MailFilterCache().reload())

    # データ受信先の監視準備
    dir_name = cfg.get_config("recv_dir", "data")
    if not os.path.exists(dir_name):
        os.makedirs(dir_name, exist_ok=True)
    from app.service.polling import add_rcv_dir

    add_rcv_dir(dir_name)

    from app.service.mail_send import start_retry_send

    # データ再送信設定準備
    smtp = loop.run_until_complete(start_retry_send())

    # バックグラウンドプロセスの実行開始
    bg_proc.append(PeriodicCallback(try_exit, timedelta(milliseconds=100)))
    for p in bg_proc:
        p.start()

    from app.service.pop import start_all_pop3

    # 全受信開始(あれば)
    loop.run_until_complete(start_all_pop3())

    # Web サーバー listen 開始
    sv = web_svr.listen(port=port_no)

    log.debug("インスタンスのリクエスト待機開始")
    IOLoop.instance().start()
